import hashlib
import hmac
import json
import logging
import socket
import time
import uuid
import warnings
from functools import cmp_to_key
from http.cookies import SimpleCookie
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union
from urllib.parse import parse_qs
from urllib.parse import unquote
from urllib.parse import urlencode
from urllib.parse import urljoin
from urllib.parse import urlparse

from jwkest import b64d
from jwkest import jwe
from jwkest import jws
from jwkest import safe_str_cmp
from jwkest.jwe import JWE
from jwkest.jwe import JWEException
from jwkest.jwe import NotSupportedAlgorithm
from jwkest.jwk import SYMKey
from jwkest.jws import NoSuitableSigningKeys
from jwkest.jws import alg2keytype
from requests import RequestException

from oic import rndstr
from oic.exception import FailedAuthentication
from oic.exception import InvalidRequest
from oic.exception import MessageException
from oic.exception import NotForMe
from oic.exception import ParameterError
from oic.exception import UnSupported
from oic.oauth2 import compact
from oic.oauth2 import error_response
from oic.oauth2 import redirect_authz_error
from oic.oauth2.base import PBase
from oic.oauth2.exception import CapabilitiesMisMatch
from oic.oauth2.exception import VerificationError
from oic.oauth2.message import Message
from oic.oauth2.message import by_schema
from oic.oauth2.provider import DELIM
from oic.oauth2.provider import STR
from oic.oauth2.provider import Endpoint
from oic.oauth2.provider import Provider as AProvider
from oic.oic import PREFERENCE2PROVIDER
from oic.oic import PROVIDER_DEFAULT
from oic.oic import Server
from oic.oic import claims_match
from oic.oic import scope2claims
from oic.oic.message import BACK_CHANNEL_LOGOUT_EVENT
from oic.oic.message import SCOPE2CLAIMS
from oic.oic.message import AccessTokenResponse
from oic.oic.message import AuthorizationResponse
from oic.oic.message import Claims
from oic.oic.message import ClientRegistrationErrorResponse
from oic.oic.message import IdToken
from oic.oic.message import OIDCMessageFactory
from oic.oic.message import OpenIDRequest
from oic.oic.message import OpenIDSchema
from oic.utils import sort_sign_alg
from oic.utils.http_util import OAUTH2_NOCACHE_HEADERS
from oic.utils.http_util import BadRequest
from oic.utils.http_util import CookieDealer
from oic.utils.http_util import Created
from oic.utils.http_util import Response
from oic.utils.http_util import SeeOther
from oic.utils.http_util import Unauthorized
from oic.utils.jwt import JWT
from oic.utils.keyio import KEYS
from oic.utils.keyio import KeyBundle
from oic.utils.keyio import KeyJar
from oic.utils.keyio import dump_jwks
from oic.utils.keyio import key_export
from oic.utils.sanitize import sanitize
from oic.utils.sdb import AccessCodeUsed
from oic.utils.sdb import ExpiredToken
from oic.utils.sdb import WrongTokenType
from oic.utils.sdb import session_get
from oic.utils.session_backend import AuthnEvent
from oic.utils.settings import OicProviderSettings
from oic.utils.settings import PyoidcSettings
from oic.utils.template_render import render_template
from oic.utils.time_util import utc_time_sans_frac

__author__ = "rohe0002"

logger = logging.getLogger(__name__)

SWD_ISSUER = "http://openid.net/specs/connect/1.0/issuer"


class InvalidRedirectURIError(Exception):
    pass


class InvalidSectorIdentifier(Exception):
    pass


class InvalidPostLogoutUri(Exception):
    """Raised when the post_logout_redirect_uris are not valid."""


def devnull(txt):
    pass


def do_authorization(user):
    return ""


def secret(seed, sid):
    msg = "{}{}{}".format(time.time(), rndstr(10), sid).encode("utf-8")
    csum = hmac.new(seed, msg, hashlib.sha224)
    return csum.hexdigest()


def code_token_response(**kwargs):
    _areq = kwargs["areq"]
    _scode = kwargs["scode"]
    _sdb = kwargs["sdb"]

    aresp = AuthorizationResponse()

    for key in ["state", "nonce", "scope"]:
        try:
            aresp[key] = _areq[key]
        except KeyError:
            pass

    aresp["code"] = _scode

    _dic = _sdb.upgrade_to_token(_scode, issue_refresh=False)
    for prop in AccessTokenResponse.c_param.keys():
        try:
            aresp[prop] = _dic[prop]
        except KeyError:
            pass

    return aresp


def location_url(response_type, redirect_uri, query):
    if response_type in [["code"], ["token"], ["none"]]:
        return "%s?%s" % (redirect_uri, query)
    else:
        return "%s#%s" % (redirect_uri, query)


def construct_uri(item):
    (base_url, query) = item
    if query:
        return "%s?%s" % (base_url, urlencode(query))
    else:
        return base_url


class AuthorizationEndpoint(Endpoint):
    etype = "authorization"
    url = "authorization"


class TokenEndpoint(Endpoint):
    etype = "token"
    url = "token"


class UserinfoEndpoint(Endpoint):
    etype = "userinfo"
    url = "userinfo"


class RegistrationEndpoint(Endpoint):
    etype = "registration"
    url = "registration"


class EndSessionEndpoint(Endpoint):
    etype = "end_session"
    url = "end_session"


RESPONSE_TYPES_SUPPORTED = [
    ["code"],
    ["token"],
    ["id_token"],
    ["code", "token"],
    ["code", "id_token"],
    ["id_token", "token"],
    ["code", "token", "id_token"],
    ["none"],
]

CAPABILITIES = {
    "response_types_supported": [" ".join(x) for x in RESPONSE_TYPES_SUPPORTED],
    "token_endpoint_auth_methods_supported": [
        "client_secret_post",
        "client_secret_basic",
        "client_secret_jwt",
        "private_key_jwt",
    ],
    "response_modes_supported": ["query", "fragment", "form_post"],
    "subject_types_supported": ["public", "pairwise"],
    "grant_types_supported": [
        "authorization_code",
        "implicit",
        "urn:ietf:params:oauth:grant-type:jwt-bearer",
        "refresh_token",
    ],
    "claim_types_supported": ["normal", "aggregated", "distributed"],
    "claims_parameter_supported": True,
    "request_parameter_supported": True,
    "request_uri_parameter_supported": True,
}


class Provider(AProvider):
    def __init__(
        self,
        name,
        sdb,
        cdb,
        authn_broker,
        userinfo,
        authz,
        client_authn,
        symkey=None,
        urlmap=None,
        keyjar=None,
        hostname="",
        template_lookup=None,
        template=None,
        verify_ssl=None,
        capabilities=None,
        schema=OpenIDSchema,
        jwks_uri="",
        jwks_name="",
        baseurl=None,
        client_cert=None,
        extra_claims=None,
        template_renderer=render_template,
        extra_scope_dict=None,
        message_factory=OIDCMessageFactory,
        post_logout_page=None,
        self_signing_alg="RS256",
        logout_path="",
        settings: Optional[PyoidcSettings] = None,
    ):
        self.settings = settings or OicProviderSettings()
        if verify_ssl is not None:
            warnings.warn(
                "`verify_ssl` is deprecated, please use `settings` instead if you need to set a non-default value.",
                DeprecationWarning,
                stacklevel=2,
            )
            self.settings.verify_ssl = verify_ssl
        if client_cert is not None:
            warnings.warn(
                "`client_cert` is deprecated, please use `settings` instead if you need to set a non-default value.",
                DeprecationWarning,
                stacklevel=2,
            )
            self.settings.client_cert = client_cert

        # This has to be defined before calling super()
        self.extra_claims = extra_claims
        self.extra_scope_dict = extra_scope_dict
        # Now we can call super()
        AProvider.__init__(
            self,
            name,
            sdb,
            cdb,
            authn_broker,
            authz,
            client_authn,
            symkey,
            urlmap,
            message_factory=message_factory,
            settings=self.settings,
        )
        # Should be a OIC Server not an OAuth2 server
        self.server = Server(
            keyjar=keyjar, message_factory=message_factory, settings=self.settings
        )
        # Same keyjar
        self.keyjar: KeyJar = self.server.keyjar

        self.endp.extend([UserinfoEndpoint, RegistrationEndpoint, EndSessionEndpoint])

        self.userinfo = userinfo
        self.template_renderer = template_renderer
        self.baseurl = baseurl or name
        self.cookie_name = "pyoidc"
        self.seed = b""
        self.sso_ttl = 0
        self.test_mode = False

        # Local filename
        self.jwks_name = jwks_name

        self.authn_as = None
        self.preferred_id_type = "public"
        self.hostname = hostname or socket.gethostname()

        self.force_jws = {"request_object": False, "id_token": False, "userinfo": False}

        self.jwx_def: Dict[str, Dict[str, str]] = {}

        self.build_jwx_def()

        self.kid: Dict[str, Dict[str, str]] = {"sig": {}, "enc": {}}

        # Allow custom schema (inheriting from OpenIDSchema) to be used -
        # additional attributes
        self.schema = schema

        # Logout connected attributes
        self.httpc = PBase(keyjar=self.keyjar, settings=self.settings)
        self.post_logout_page = post_logout_page
        self.signing_alg = self_signing_alg
        self.logout_path = logout_path
        self.logout_verify_url = ""

    @property
    def default_capabilities(self):
        """Define default capabilities for implementation."""
        return CAPABILITIES

    def build_jwx_def(self):
        self.jwx_def = {}

        for _typ in ["signing_alg", "encryption_alg", "encryption_enc"]:
            self.jwx_def[_typ] = {}
            for item in ["id_token", "userinfo"]:
                cap_param = "{}_{}_values_supported".format(item, _typ)
                try:
                    self.jwx_def[_typ][item] = self.capabilities[cap_param][0]
                except KeyError:
                    self.jwx_def[_typ][item] = ""

    def set_mode(self, mode):
        """
        Prepare OP based on parameters that govern how this OP will behave.

        :param mode:
        :return:
        """
        # Is there a signing algorithm I should use
        try:
            self.jwx_def["signing_alg"]["id_token"] = mode["sign"]
            self.jwx_def["signing_alg"]["userinfo"] = mode["sign"]
        except KeyError:
            pass
        else:
            # make sure id_token_signed_response_alg is set in client register
            # response. This will make it happen in match_preferences()
            for val in PREFERENCE2PROVIDER.values():
                if val.endswith("signing_alg_values_supported"):
                    self.capabilities[val] = [mode["sign"]]

        # Is there a encryption algorithm I should use
        try:
            _enc_alg = mode["enc_alg"]
        except KeyError:
            pass
        else:
            # make sure id_token_signed_response_alg is set in client register
            # response. This will make it happen in match_preferences()
            for val in PREFERENCE2PROVIDER.values():
                if val.endswith("encryption_alg_values_supported"):
                    self.capabilities[val] = [_enc_alg]

        # Is there a encryption enc algorithm I should use
        try:
            _enc_enc = mode["enc_enc"]
        except KeyError:
            pass
        else:
            # make sure id_token_signed_response_alg is set in client register
            # response. This will make it happen in match_preferences()
            for val in PREFERENCE2PROVIDER.values():
                if val.endswith("encryption_enc_values_supported"):
                    self.capabilities[val] = [_enc_enc]

    def id_token_as_signed_jwt(
        self,
        session,
        loa="2",
        alg="",
        code=None,
        access_token=None,
        user_info=None,
        auth_time=0,
        exp=None,
        extra_claims=None,
        **kwargs,
    ):

        if alg == "":
            alg = self.jwx_def["signing_alg"]["id_token"]

        if alg:
            logger.debug("Signing alg: %s [%s]" % (alg, alg2keytype(alg)))
        else:
            alg = "none"

        _idt = self.server.make_id_token(
            session,
            loa,
            self.name,
            alg,
            code,
            access_token,
            user_info,
            auth_time,
            exp,
            extra_claims,
        )

        try:
            ckey = kwargs["keys"]
        except KeyError:
            try:
                _keyjar = kwargs["keyjar"]
            except KeyError:
                _keyjar = self.keyjar

            logger.debug("id_token: %s" % sanitize(_idt.to_dict()))
            # My signing key if its RS*, can use client secret if HS*
            if alg.startswith("HS"):
                logger.debug("client_id: %s" % session["client_id"])
                ckey = _keyjar.get_signing_key(alg2keytype(alg), session["client_id"])
                if not ckey:  # create a new key
                    _secret = self.cdb[session["client_id"]]["client_secret"]
                    ckey = [SYMKey(key=_secret)]
            else:
                if "" in self.keyjar:
                    ckey = _keyjar.get_signing_key(alg2keytype(alg), "", alg=alg)
                else:
                    ckey = None

        _signed_jwt = _idt.to_jwt(key=ckey, algorithm=alg)

        return _signed_jwt

    def _parse_openid_request(self, request, **kwargs):
        return OpenIDRequest().from_jwt(request, keyjar=self.keyjar, **kwargs)

    def _parse_id_token(self, id_token, redirect_uri):
        try:
            return IdToken().from_jwt(id_token, keyjar=self.keyjar)
        except Exception as err:
            logger.error("Faulty id_token: %s" % id_token)
            logger.error("Exception: %s" % (err.__class__.__name__,))
            id_token = IdToken().from_jwt(id_token, verify=False)
            logger.error("IdToken: %s" % id_token.to_dict())
            return redirect_authz_error("invalid_id_token_object", redirect_uri)

    @staticmethod
    def get_sector_id(redirect_uri, client_info):
        """
        Pick the sector id given a number of factors.

        :param redirect_uri: The redirect_uri used
        :param client_info: Information provided by the client in the client registration
        :return: A sector_id or None
        """
        _redirect_uri = unquote(redirect_uri)

        part = urlparse(_redirect_uri)
        if part.fragment:
            raise ValueError
        _base = part._replace(query="").geturl()

        sid = ""
        try:
            if _base in client_info["si_redirects"]:
                sid = client_info["sector_id"]
        except KeyError:
            try:
                uit = client_info["subject_type"]
                if uit == "pairwise":
                    sid = _base
            except KeyError:
                pass

        return sid

    def _verify_client(self, areq, aud):
        if areq["client_id"] in aud:
            return True
        else:
            return False

    def required_user(self, areq):
        req_user = ""
        try:
            _req = areq["request"]
        except KeyError:
            _req = areq

        if "id_token_hint" in _req:
            try:
                req_user = _req["id_token_hint"]["sub"]
                aud = _req["id_token_hint"]["aud"]
            except (KeyError, TypeError):
                # A signed jwt, should verify signature if I can
                jso = json.loads(b64d(str(_req["id_token_hint"].split(".")[1])))
                req_user = jso["sub"]
                aud = jso["aud"]

            if not self._verify_client(areq, aud):
                req_user = ""

        return req_user

    def is_session_revoked(self, request="", cookie=None):
        areq = parse_qs(request)
        authn, _ = self.pick_auth(areq)
        identity, _ts = authn.authenticated_as(cookie)
        return self.sdb.is_revoke_uid(identity["uid"])

    def verify_endpoint(self, request="", cookie=None, **kwargs):
        """
        Verify endpoint.

        :param request:
        :param cookie:
        :param kwargs:
        :return:
        """
        logger.debug("verify request: %s" % sanitize(request))

        if isinstance(request, dict):
            _req = request
        else:
            _req = compact(parse_qs(request))

        try:
            areq = Message().from_urlencoded(_req["query"])
        except KeyError:
            areq = _req

        logger.debug("REQ: %s", sanitize(areq))
        try:
            authn, acr = self.pick_auth(areq, "exact")
        except Exception as err:
            logger.exception("%s", sanitize(err))
            raise

        kwargs["cookie"] = cookie
        return authn.verify(request=_req, **kwargs)

    def setup_session(self, areq, authn_event, cinfo):
        try:
            oidc_req = areq["request"]
        except KeyError:
            oidc_req = None

        sid = self.sdb.create_authz_session(authn_event, areq, oidreq=oidc_req)
        kwargs = {}
        for param in ["sector_id", "subject_type"]:
            try:
                kwargs[param] = cinfo[param]
            except KeyError:
                pass

        self.sdb.do_sub(sid, cinfo["client_salt"], **kwargs)
        return sid

    def match_sp_sep(self, first, second):
        one = [set(v.split(" ")) for v in first]
        other = [set(v.split(" ")) for v in second]
        if not any(rt in one for rt in other):
            return False
        return True

    def filter_request(self, req):
        _cap = self.capabilities

        before = req.to_dict()

        if "claims" in req:
            if _cap["claims_parameter_supported"]:
                if _cap["claims_supported"]:
                    for part in ["userinfo", "id_token"]:
                        if part in req["claims"]:
                            _keys = list(req["claims"][part].keys())
                            for c in _keys:
                                if c not in _cap["claims_supported"]:
                                    del req["claims"][part][c]
            else:
                del req["claims"]

        if "scope" in req:
            _scopes = [s for s in req["scope"] if s in _cap["scopes_supported"]]
            req["scope"] = _scopes

        if "request" in req:
            if _cap["request_parameter_supported"] is False:
                raise InvalidRequest("Contains unsupported request parameter")

        if "request_uri" in req:
            if _cap["request_uri_parameter_supported"] is False:
                raise InvalidRequest("Contains unsupported request parameter")

        if "response_mode" in req:
            if req["response_mode"] not in _cap["response_modes_supported"]:
                raise InvalidRequest("Contains unsupported response mode")

        if "response_type" in req:
            if not self.match_sp_sep(
                [" ".join(req["response_type"])], _cap["response_types_supported"]
            ):
                raise InvalidRequest("Contains unsupported response type")

        if before != req.to_dict():
            msg = "Request modified from %s to %s"
            logger.warning(msg, before, req.to_dict())

        return req

    def auth_init(self, request):
        """Overriden since the filter_request can throw an InvalidRequest."""
        try:
            return super().auth_init(request)
        except InvalidRequest as err:
            return error_response("invalid_request", "%s" % err)

    def authorization_endpoint(self, request="", cookie=None, **kwargs):
        """
        Authorize the client.

        :param request: The client request
        """
        info = self.auth_init(request)
        if isinstance(info, Response):
            return info

        areq = info["areq"]
        logger.info("authorization_request: %s" % (sanitize(areq.to_dict()),))

        _cid = areq["client_id"]
        cinfo = self.cdb[str(_cid)]
        if _cid not in self.keyjar.issuer_keys:
            self.recuperate_keys(_cid, cinfo)

        req_user = self.required_user(areq)
        if req_user:
            sids = self.sdb.get__by_sub(req_user)
            if sids:
                # anyone will do
                authn_event = self.sdb.get_authentication_event(sids[-1])
                # Is the authentication event to be regarded as valid ?
                if authn_event.valid():
                    sid = self.setup_session(areq, authn_event, cinfo)
                    return self.authz_part2(authn_event.uid, areq, sid, cookie=cookie)

            kwargs["req_user"] = req_user

        authnres = self.do_auth(
            info["areq"], info["redirect_uri"], cinfo, request, cookie, **kwargs
        )

        if isinstance(authnres, Response):
            return authnres

        logger.debug("- authenticated -")
        logger.debug("AREQ keys: %s", list(areq.keys()))

        sid = self.setup_session(areq, authnres["authn_event"], cinfo)
        return self.authz_part2(authnres["user"], areq, sid, cookie=cookie)

    def authz_part2(self, user, areq, sid, **kwargs):
        result = self._complete_authz(user, areq, sid, **kwargs)
        if isinstance(result, Response):
            return result
        else:
            aresp, headers, redirect_uri, fragment_enc = result

        if "check_session_iframe" in self.capabilities:
            salt = rndstr()
            authn_event = self.sdb.get_authentication_event(sid)  # use the last session
            state = str(authn_event.authn_time)
            aresp["session_state"] = self._compute_session_state(
                state, salt, areq["client_id"], redirect_uri
            )
            headers.append(
                self.write_session_cookie(state, http_only=False, same_site="None")
            )

        # as per the mix-up draft don't add iss and client_id if they are
        # already in the id_token.
        if "id_token" not in aresp:
            aresp["iss"] = self.name

        aresp["client_id"] = areq["client_id"]

        if self.events:
            self.events.store("protocol response", aresp)

        response = sanitize(aresp.to_dict())
        logger.info("authorization response: %s", response)

        location = aresp.request(redirect_uri, fragment_enc)
        msg = "Redirected to: '%s' :: %s"
        logger.debug(msg, sanitize(location), type(location))

        return SeeOther(str(location), headers=headers)

    def userinfo_in_id_token_claims(self, session):
        """
        Put userinfo claims in the id token.

        :param session:
        :return:
        """
        itc = self.server.id_token_claims(session)
        if not itc:
            return None

        _claims = by_schema(self.schema, **itc)

        if _claims:
            return self._collect_user_info(session, _claims)
        else:
            return None

    def recuperate_keys(self, cid: str, client_info: Dict[str, str]) -> None:
        """Try to recuperate lost keys."""
        msg = "Lost keys for %s trying to recuperate!"
        logger.warning(msg, cid)

        self.keyjar.issuer_keys[cid] = []
        # Add client secret as a symmetric key
        self.keyjar.add_symmetric(
            cid, client_info["client_secret"], usage=["enc", "sig"]
        )
        # Try to renew from jwks or jwks_uri
        if client_info.get("jwks_uri") is not None:
            self.keyjar.add(cid, client_info["jwks_uri"])
        elif client_info.get("jwks") is not None:
            self.keyjar.import_jwks(client_info["jwks"], cid)
        else:
            logger.warning("No keys to recover.")

    def encrypt(self, payload, client_info, cid, val_type="id_token", cty=""):
        """
        Handle the encryption of a payload.

        Shouldn't get here unless there are encrypt parameters in client info

        :param payload: The information to be encrypted
        :param client_info: Client information
        :param cid: Client id
        :return: The encrypted information as a JWT
        """
        try:
            alg = client_info["%s_encrypted_response_alg" % val_type]
        except KeyError:
            logger.warning("%s NOT defined means no encryption", val_type)
            return payload
        else:
            try:
                enc = client_info["%s_encrypted_response_enc" % val_type]
            except KeyError as err:  # if not defined-> A128CBC-HS256 (default)
                logger.warning("undefined parameter: %s", err)
                logger.info("using default")
                enc = "A128CBC-HS256"

        logger.debug("alg=%s, enc=%s, val_type=%s" % (alg, enc, val_type))
        if cid not in self.keyjar:
            self.recuperate_keys(cid, client_info)
        keys = self.keyjar.get_encrypt_key(owner=cid)
        kwargs = {"alg": alg, "enc": enc}
        if cty:
            kwargs["cty"] = cty

        # use the clients public key for encryption
        _jwe = JWE(payload, **kwargs)
        return _jwe.encrypt(keys, context="public")

    def sign_encrypt_id_token(
        self, sinfo, client_info, areq, code=None, access_token=None, user_info=None
    ):
        """
        Sign and or encrypt a IDToken.

        :param sinfo: Session information
        :param client_info: Client information
        :param areq: The request
        :param code: Access grant
        :param access_token: Access Token
        :param user_info: User information
        :return: IDToken instance
        """
        try:
            alg = client_info["id_token_signed_response_alg"]
        except KeyError:
            try:
                alg = self.jwx_def["signing_alg"]["id_token"]
            except KeyError:
                alg = PROVIDER_DEFAULT["id_token_signed_response_alg"]
            else:
                if not alg:
                    alg = PROVIDER_DEFAULT["id_token_signed_response_alg"]

        _authn_event = AuthnEvent.from_json(sinfo["authn_event"])
        id_token = self.id_token_as_signed_jwt(
            sinfo,
            loa=_authn_event.authn_info,
            alg=alg,
            code=code,
            access_token=access_token,
            user_info=user_info,
            auth_time=_authn_event.authn_time,
        )

        # Then encrypt
        if "id_token_encrypted_response_alg" in client_info:
            id_token = self.encrypt(
                id_token, client_info, areq["client_id"], "id_token", "JWT"
            )

        return id_token

    def code_grant_type(self, areq):
        """
        Token authorization using Code Grant.

        RFC6749 section 4.1
        """
        _sdb = self.sdb
        _log_debug = logger.debug

        client_info = self.cdb[str(areq["client_id"])]

        try:
            _access_code = areq["code"].replace(" ", "+")
        except KeyError:  # Missing code parameter - absolutely fatal
            return error_response("invalid_request", descr="Missing code")

        # assert that the code is valid
        if self.sdb.is_revoked(_access_code):
            return error_response("invalid_request", descr="Token is revoked")

        # Session might not exist or _access_code malformed
        try:
            _info = _sdb[_access_code]
        except KeyError:
            return error_response("invalid_request", descr="Code is invalid")

        # If redirect_uri was in the initial authorization request verify that it is here as well
        # Mismatch would raise in oic.oauth2.provider.Provider.token_endpoint
        if "redirect_uri" in _info and "redirect_uri" not in areq:
            return error_response("invalid_request", descr="Missing redirect_uri")

        _log_debug("All checks OK")

        issue_refresh = False
        permissions = _info.get("permission", ["offline_access"]) or ["offline_access"]
        if "offline_access" in _info["scope"] and "offline_access" in permissions:
            issue_refresh = True

        try:
            _tinfo = _sdb.upgrade_to_token(_access_code, issue_refresh=issue_refresh)
        except AccessCodeUsed as err:
            logger.error("%s" % err)
            # Should revoke the token issued to this access code
            _sdb.revoke_all_tokens(_access_code)
            return error_response("access_denied", descr="Access Code already used")

        if "openid" in _info["scope"]:
            userinfo = self.userinfo_in_id_token_claims(_info)
            try:
                _idtoken = self.sign_encrypt_id_token(
                    _info, client_info, areq, user_info=userinfo
                )
            except (JWEException, NoSuitableSigningKeys) as err:
                logger.warning(str(err))
                return error_response(
                    "invalid_request", descr="Could not sign/encrypt id_token"
                )

            _sdb.update_by_token(_access_code, "id_token", _idtoken)

        # Refresh the _tinfo
        _tinfo = _sdb[_access_code]

        _log_debug("_tinfo: %s" % sanitize(_tinfo))

        response_cls = self.server.message_factory.get_response_type("token_endpoint")
        atr = response_cls(**by_schema(response_cls, **_tinfo))

        logger.info("access_token_response: %s" % sanitize(atr.to_dict()))

        return Response(
            atr.to_json(), content="application/json", headers=OAUTH2_NOCACHE_HEADERS
        )

    def refresh_token_grant_type(self, areq):
        """
        Token refresh.

        RFC6749 section 6
        """
        _sdb = self.sdb
        _log_debug = logger.debug

        client_id = str(areq["client_id"])
        client_info = self.cdb[client_id]

        rtoken = areq["refresh_token"]
        try:
            _info = _sdb.refresh_token(rtoken, client_id=client_id)
        except ExpiredToken:
            return error_response("invalid_request", descr="Refresh token is expired")
        except WrongTokenType:
            return error_response("invalid_request", descr="Not a refresh token")

        if "openid" in _info["scope"] and "authn_event" in _info:
            userinfo = self.userinfo_in_id_token_claims(_info)
            try:
                _idtoken = self.sign_encrypt_id_token(
                    _info, client_info, areq, user_info=userinfo
                )
            except (JWEException, NoSuitableSigningKeys) as err:
                logger.warning(str(err))
                return error_response(
                    "invalid_request", descr="Could not sign/encrypt id_token"
                )

            sid = _sdb.access_token.get_key(_info["access_token"])
            _sdb.update(sid, "id_token", _idtoken)

        _log_debug("_info: %s" % sanitize(_info))

        response_cls = self.server.message_factory.get_response_type("token_endpoint")
        atr = response_cls(**by_schema(response_cls, **_info))

        logger.info("access_token_response: %s" % sanitize(atr.to_dict()))

        return Response(
            atr.to_json(), content="application/json", headers=OAUTH2_NOCACHE_HEADERS
        )

    def client_credentials_grant_type(self, areq):
        """
        Token authorization using client credentials.

        RFC6749 section 4.4
        """
        # Not supported in OpenID Connect
        return error_response("unsupported_grant_type", descr="Unsupported grant_type")

    def password_grant_type(self, areq):
        """
        Token authorization using Resource owner password credentials.

        RFC6749 section 4.3
        """
        # Not supported in OpenID Connect
        return error_response("unsupported_grant_type", descr="Unsupported grant_type")

    def _collect_user_info(self, session, userinfo_claims=None):
        """
        Collect information about a user.

        This can happen in two cases, either when constructing an IdToken or
        when returning user info through the UserInfo endpoint

        :param session: Session information
        :param userinfo_claims: user info claims
        :return: User info
        """
        if userinfo_claims is None:
            uic = scope2claims(session["scope"], extra_scope_dict=self.extra_scope_dict)

            # Get only keys allowed by user and update the dict if such info
            # is stored in session
            perm_set = session.get("permission")
            if perm_set:
                uic = {key: uic[key] for key in uic if key in perm_set}

            if "oidreq" in session:
                uic = self.server.update_claims(session, "oidreq", "userinfo", uic)
            else:
                uic = self.server.update_claims(session, "authzreq", "userinfo", uic)
            if uic:
                userinfo_claims = Claims(**uic)
            else:
                userinfo_claims = None

            logger.debug("userinfo_claim: %s" % sanitize(userinfo_claims.to_dict()))

        logger.debug("Session info: %s" % sanitize(session))

        if "authn_event" in session:
            uid = AuthnEvent.from_json(session["authn_event"]).uid
        else:
            uid = session["uid"]

        info = self.userinfo(uid, session["client_id"], userinfo_claims)

        if "sub" in userinfo_claims:
            if not claims_match(session["sub"], userinfo_claims["sub"]):
                raise FailedAuthentication("Unmatched sub claim")

        info["sub"] = session["sub"]
        logger.debug("user_info_response: %s", info)

        return info

    def signed_userinfo(self, client_info, userinfo, session):
        """
        Create a JWS with the userinfo as payload.

        :param client_info: Client registration information
        :param userinfo: An OpenIDSchema instance
        :param session: Session information
        :return: A JWS containing the userinfo as a JWT
        """
        try:
            algo = client_info["userinfo_signed_response_alg"]
        except KeyError:  # Fall back to default
            algo = self.jwx_def["signing_alg"]["userinfo"]

        if algo == "none":
            key: List[KEYS] = []
        else:
            if algo.startswith("HS"):
                key = self.keyjar.get_signing_key(
                    alg2keytype(algo), client_info["client_id"], alg=algo
                )
            else:
                # Use my key for signing
                key = self.keyjar.get_signing_key(alg2keytype(algo), "", alg=algo)
            if not key:
                return error_response("invalid_request", descr="Missing signing key")

        jinfo = userinfo.to_jwt(key, algo)
        if "userinfo_encrypted_response_alg" in client_info:
            # encrypt with clients public key
            jinfo = self.encrypt(
                jinfo, client_info, session["client_id"], "userinfo", "JWT"
            )
        return jinfo

    def userinfo_endpoint(self, request="", **kwargs):
        """
        Endpoint for collecting the UserInfo.

        :param request: The request in a string format or as a dictionary
        """
        logger.debug("userinfo_endpoint: request={}, kwargs={}".format(request, kwargs))

        try:
            _token = self._parse_access_token(request, **kwargs)
        except ParameterError:
            return error_response("invalid_request", descr="Token is malformed")
        return self._do_user_info(_token, **kwargs)

    def _parse_access_token(self, request, **kwargs):
        if not request or "access_token" not in request:
            _token = kwargs.get("authn", "") or ""
            if not _token.startswith("Bearer "):
                raise ParameterError("Token is missing or malformed")
            _token = _token[len("Bearer ") :]
            logger.debug("Bearer token {} chars".format(len(_token)))
        else:
            args = {"data": request}
            if isinstance(request, dict):
                args["sformat"] = "dict"
            uireq = self.server.parse_user_info_request(**args)
            logger.debug("user_info_request: %s" % sanitize(uireq))
            _token = uireq["access_token"].replace(" ", "+")

        return _token

    def _do_user_info(self, token, **kwargs):
        try:
            _log_debug = kwargs["logger"].debug
        except KeyError:
            _log_debug = logger.debug

        _sdb = self.sdb
        # should be an access token
        try:
            typ, key = _sdb.access_token.type_and_key(token)
        except Exception:
            return error_response(
                "invalid_token", descr="Invalid Token", status_code=401
            )

        _log_debug("access_token type: '%s'" % (typ,))

        if typ != "T":
            logger.error("Wrong token type: {}".format(typ))
            raise FailedAuthentication("Wrong type of token")

        if _sdb.access_token.is_expired(token):
            return error_response(
                "invalid_token", descr="Token is expired", status_code=401
            )

        if _sdb.is_revoked(key):
            return error_response(
                "invalid_token", descr="Token is revoked", status_code=401
            )
        session = _sdb[key]

        # Scope can translate to userinfo_claims

        info = self.schema(**self._collect_user_info(session))

        # Should I return a JSON or a JWT ?
        _cinfo = self.cdb.get(session["client_id"])
        if _cinfo is None:
            return error_response("unauthorized_client", descr="Unknown client")
        try:
            if "userinfo_signed_response_alg" in _cinfo:
                # Will also encrypt if defined in cinfo
                jinfo = self.signed_userinfo(_cinfo, info, session)
                content_type = "application/jwt"
            elif "userinfo_encrypted_response_alg" in _cinfo:
                jinfo = info.to_json()
                jinfo = self.encrypt(
                    jinfo, _cinfo, session["client_id"], "userinfo", ""
                )
                content_type = "application/jwt"
            else:
                jinfo = info.to_json()
                content_type = "application/json"
        except NotSupportedAlgorithm as err:
            return error_response(
                "invalid_request",
                descr="Not supported algorithm: {}".format(err.args[0]),
            )
        except JWEException:
            return error_response("invalid_request", descr="Could not encrypt")

        return Response(jinfo, content=content_type)

    def check_session_endpoint(self, request, **kwargs):
        try:
            _log_info = kwargs["logger"].info
        except KeyError:
            _log_info = logger.info

        if not request:
            _tok = kwargs["authn"]
            if not _tok:
                return error_response("invalid_request", descr="Illegal token")

        if self.test_mode:
            _log_info("check_session_request: %s" % sanitize(request))
        idt = self.server.parse_check_session_request(query=request)
        if self.test_mode:
            _log_info("check_session_response: %s" % idt.to_dict())

        return Response(idt.to_json(), content="application/json")

    @staticmethod
    def _verify_url(url, urlset):
        part = urlparse(url)

        for reg, _ in urlset:
            _part = urlparse(reg)
            if part.scheme == _part.scheme and part.netloc == _part.netloc:
                return True

        return False

    def match_client_request(self, request):
        for _pref, _prov in PREFERENCE2PROVIDER.items():
            if _pref in request:
                if _pref == "response_types":
                    if not self.match_sp_sep(request[_pref], self.capabilities[_prov]):
                        raise CapabilitiesMisMatch(_pref)
                else:
                    if isinstance(request[_pref], str):
                        if request[_pref] not in self.capabilities[_prov]:
                            raise CapabilitiesMisMatch(_pref)
                    else:
                        if not set(request[_pref]).issubset(
                            set(self.capabilities[_prov])
                        ):
                            raise CapabilitiesMisMatch(_pref)

    def do_client_registration(self, request, client_id, ignore=None):
        if ignore is None:
            ignore = []

        _cinfo = self.cdb[client_id].copy()
        logger.debug("_cinfo: %s" % sanitize(_cinfo))

        for key, val in request.items():
            if key not in ignore:
                _cinfo[key] = val

        if "post_logout_redirect_uris" in request:
            try:
                plruri = self._verify_post_logout_uri(request)
            except InvalidPostLogoutUri as err:
                error = ClientRegistrationErrorResponse(
                    error="invalid_configuration_parameter", error_description=str(err)
                )
                return Response(
                    error.to_json(),
                    content="application/json",
                    status="400 Bad Request",
                )
            _cinfo["post_logout_redirect_uris"] = plruri

        if "redirect_uris" in request:
            try:
                ruri = self.verify_redirect_uris(request)
                _cinfo["redirect_uris"] = ruri
            except InvalidRedirectURIError as e:
                error = ClientRegistrationErrorResponse(
                    error="invalid_redirect_uri", error_description=str(e)
                )
                return Response(
                    error.to_json(), content="application/json", status_code=400
                )

        if "sector_identifier_uri" in request:
            try:
                (
                    _cinfo["si_redirects"],
                    _cinfo["sector_id"],
                ) = self._verify_sector_identifier(request)
            except InvalidSectorIdentifier as err:
                return error_response("invalid_configuration_parameter", descr=str(err))
        elif "redirect_uris" in request and len(request["redirect_uris"]) > 1:
            # check that the hostnames are the same
            host = ""
            for url in request["redirect_uris"]:
                part = urlparse(url)
                _host = part.netloc.split(":")[0]
                if not host:
                    host = _host
                else:
                    if host != _host:
                        return error_response(
                            "invalid_configuration_parameter",
                            descr="'sector_identifier_uri' must be registered",
                        )

        for item in ["policy_uri", "logo_uri", "tos_uri"]:
            if item in request:
                if self._verify_url(request[item], _cinfo["redirect_uris"]):
                    _cinfo[item] = request[item]
                else:
                    return error_response(
                        "invalid_configuration_parameter",
                        descr="%s pointed to illegal URL" % item,
                    )

        # Do I have the necessary keys
        for item in ["id_token_signed_response_alg", "userinfo_signed_response_alg"]:
            if item in request:
                if request[item] in self.capabilities[PREFERENCE2PROVIDER[item]]:
                    ktyp = jws.alg2keytype(request[item])
                    # do I have this ktyp and for EC type keys the curve
                    if ktyp not in ["none", "oct"]:
                        _k = self.keyjar.get_signing_key(ktyp, alg=request[item])
                        if not _k:
                            del _cinfo[item]

        try:
            self.keyjar.load_keys(request, client_id)
            try:
                n_keys = len(self.keyjar[client_id])
                msg = "found {} keys for client_id={}"
                logger.debug(msg.format(n_keys, client_id))
            except KeyError:
                pass
        except Exception as err:
            logger.error("Failed to load client keys: %s" % sanitize(request.to_dict()))
            logger.error("%s", err)
            logger.debug("Verify SSL: {}".format(self.keyjar.verify_ssl))
            error = ClientRegistrationErrorResponse(
                error="invalid_configuration_parameter", error_description="%s" % err
            )
            return Response(
                error.to_json(), content="application/json", status="400 Bad Request"
            )

        return _cinfo

    @staticmethod
    def verify_redirect_uris(registration_request):
        verified_redirect_uris = []
        try:
            client_type = registration_request["application_type"]
        except KeyError:  # default
            client_type = "web"

        if client_type == "web":
            try:
                if registration_request["response_types"] == ["code"]:
                    must_https = False
                else:  # one has to be implicit or hybrid
                    must_https = True
            except KeyError:
                must_https = True
        else:
            must_https = False

        for uri in registration_request["redirect_uris"]:
            p = urlparse(uri)
            if client_type == "native":
                if p.scheme not in ["http", "https"]:  # Custom scheme
                    pass
                elif p.scheme == "http" and p.hostname in ["localhost", "127.0.0.1"]:
                    pass
                else:
                    logger.error(
                        "InvalidRedirectURI: scheme:%s, hostname:%s",
                        p.scheme,
                        p.hostname,
                    )
                    raise InvalidRedirectURIError(
                        "Redirect_uri must use custom scheme or http and localhost"
                    )
            elif must_https and p.scheme != "https":
                raise InvalidRedirectURIError("None https redirect_uri not allowed")
            elif p.fragment:
                raise InvalidRedirectURIError("redirect_uri contains fragment")

            query = p.query if p.query else None
            base = p._replace(query="").geturl()
            if query:
                verified_redirect_uris.append((base, parse_qs(query)))
            else:
                verified_redirect_uris.append((base, query))

        return verified_redirect_uris

    def _verify_post_logout_uri(self, request):
        """Verify correct format of post_logout_redirect_uris."""
        plruri = []
        for uri in request["post_logout_redirect_uris"]:
            part = urlparse(uri)
            if part.fragment:
                raise InvalidPostLogoutUri(
                    "post_logout_redirect_uris contains fragment"
                )
            query = part.query if part.query else None
            base = part._replace(query="").geturl()
            if query:
                plruri.append((base, parse_qs(query)))
            else:
                plruri.append((base, query))
        return plruri

    def _verify_sector_identifier(self, request):
        """
        Verify `sector_identifier_uri` is reachable and that it contains `redirect_uri`s.

        :param request: Provider registration request
        :return: si_redirects, sector_id
        :raises: InvalidSectorIdentifier
        """
        si_url = request["sector_identifier_uri"]
        try:
            res = self.server.http_request(si_url)
        except RequestException as err:
            logger.error(err)
            res = None

        if not res:
            raise InvalidSectorIdentifier("Couldn't open sector_identifier_uri")

        logger.debug("sector_identifier_uri => %s", sanitize(res.text))

        try:
            si_redirects = json.loads(res.text)
        except ValueError:
            raise InvalidSectorIdentifier(
                "Error deserializing sector_identifier_uri content"
            )

        if "redirect_uris" in request:
            logger.debug("redirect_uris: %s", request["redirect_uris"])
            for uri in request["redirect_uris"]:
                if uri not in si_redirects:
                    raise InvalidSectorIdentifier(
                        "redirect_uri missing from sector_identifiers"
                    )

        return si_redirects, si_url

    @staticmethod
    def comb_uri(args):
        for param in ["redirect_uris", "post_logout_redirect_uris"]:
            if param not in args:
                continue

            val = []
            for base, query_dict in args[param]:
                if query_dict:
                    query_string = urlencode(
                        [(key, v) for key in query_dict for v in query_dict[key]]
                    )
                    val.append("%s?%s" % (base, query_string))
                else:
                    val.append(base)

            args[param] = val

    def create_registration(self, authn=None, request=None, **kwargs):
        logger.debug("@registration_endpoint: <<%s>>" % sanitize(request))

        request_cls = self.server.message_factory.get_request_type(
            "registration_endpoint"
        )
        try:
            request = request_cls().deserialize(request, "json")
        except MessageException:
            request = request_cls().deserialize(request)

        logger.info("registration_request:%s" % sanitize(request.to_dict()))

        result = self.client_registration_setup(request)
        if isinstance(result, Response):
            return result

        return Created(
            result.to_json(),
            content="application/json",
            headers=[("Cache-Control", "no-store")],
        )

    @staticmethod
    def client_secret_expiration_time():
        """
        Return client_secret expiration time.

        Split for easy customization.
        """
        return utc_time_sans_frac() + 86400

    def client_registration_setup(self, request):
        try:
            request.verify()
        except MessageException as err:
            if "type" not in request:
                return error_response("invalid_type", descr="%s" % err)
            else:
                return error_response(
                    "invalid_configuration_parameter", descr="%s" % err
                )

        request.rm_blanks()
        try:
            self.match_client_request(request)
        except CapabilitiesMisMatch as err:
            return error_response(
                "invalid_request", descr="Don't support proposed %s" % err
            )

        # create new id och secret
        client_id = rndstr(12)
        while client_id in self.cdb:
            client_id = rndstr(12)

        client_secret = secret(self.seed, client_id)

        _rat = rndstr(32)
        reg_enp = ""
        for endp in self.endp:
            if endp.etype == "registration":
                reg_enp = urljoin(self.baseurl, endp.url)
                break

        self.cdb[client_id] = {
            "client_id": client_id,
            "client_secret": client_secret,
            "registration_access_token": _rat,
            "registration_client_uri": "%s?client_id=%s" % (reg_enp, client_id),
            "client_secret_expires_at": self.client_secret_expiration_time(),
            "client_id_issued_at": utc_time_sans_frac(),
            "client_salt": rndstr(8),
        }

        _cinfo = self.do_client_registration(
            request,
            client_id,
            ignore=["redirect_uris", "policy_uri", "logo_uri", "tos_uri"],
        )
        if isinstance(_cinfo, Response):
            return _cinfo

        response_cls = self.server.message_factory.get_response_type(
            "registration_endpoint"
        )
        args = dict([(k, v) for k, v in _cinfo.items() if k in response_cls.c_param])

        self.comb_uri(args)
        response = response_cls(**args)

        # Add the client_secret as a symmetric key to the keyjar
        if client_secret:
            self.keyjar.add_symmetric(client_id, str(client_secret))

        self.cdb[client_id] = _cinfo

        try:
            self.cdb.sync()
        except AttributeError:  # Not all databases can be sync'ed
            pass

        logger.info("registration_response: %s" % sanitize(response.to_dict()))

        return response

    def registration_endpoint(self, request, authn=None, method="POST", **kwargs):
        if method.lower() == "post":
            return self.create_registration(authn, request, **kwargs)
        elif method.lower() == "get":
            return self.read_registration(authn, request, **kwargs)
        elif method.lower() == "put":
            return self.alter_registration(authn, request, **kwargs)
        elif method.lower() == "delete":
            return self.delete_registration(authn, request, **kwargs)
        return error_response("Unsupported method", descr="Unsupported HTTP method")

    def read_registration(self, authn, request, **kwargs):
        """
        Read all information this server has on a client.

        Authorization is done by using the access token that was return as
        part of the client registration result.

        :param authn: The Authorization HTTP header
        :param request: The query part of the URL
        :param kwargs: Any other arguments
        :return:
        """
        logger.debug("authn: %s, request: %s" % (sanitize(authn), sanitize(request)))

        # verify the access token, has to be key into the client information
        # database.
        if not authn.startswith("Bearer "):
            return error_response("invalid_request")
        token = authn[len("Bearer ") :]

        # Get client_id from request
        _info = parse_qs(request)
        cid = _info.get("client_id")
        if cid is None:
            return Unauthorized()
        client_id = cid[0]

        cdb_entry = self.cdb.get(client_id)
        if cdb_entry is None:
            return Unauthorized()
        reg_token = cdb_entry.get("registration_access_token", "")
        if not safe_str_cmp(reg_token, token):
            return Unauthorized()

        logger.debug("Client '%s' reads client info" % client_id)
        response_cls = self.server.message_factory.get_response_type(
            "registration_endpoint"
        )
        args = dict(
            [
                (k, v)
                for k, v in self.cdb[client_id].items()
                if k in response_cls.c_param
            ]
        )

        self.comb_uri(args)
        response = response_cls(**args)

        return Response(
            response.to_json(),
            content="application/json",
            headers=[("Cache-Control", "no-store")],
        )

    def alter_registration(self, authn, request, **kwargs):
        """
        Alter the client info on server side.

        :param authn: Authorization HTTP header
        :param request: Query part of the request
        :return: Response with updated client info
        """
        return error_response(
            "Unsupported operation",
            descr="Altering of the registration is not supported",
            status_code=403,
        )

    def delete_registration(self, authn, request, **kwargs):
        """
        Delete the client info on server side.

        :param authn: Authorization HTTP header
        :param request: Query part of the request
        :return: Response with updated client info
        """
        return error_response(
            "Unsupported operation",
            descr="Deletion of the registration is not supported",
            status_code=403,
        )

    def provider_features(self, provider_config=None):
        """
        Specify what the server capabilities are.

        :return: ProviderConfigurationResponse instance
        """
        _provider_info = super().provider_features(provider_config=provider_config)

        # Parse scopes - override the base class
        _scopes = list(SCOPE2CLAIMS.keys())
        if self.extra_scope_dict is not None:
            _scopes.extend(self.extra_scope_dict.keys())
        # Remove duplicates if any
        _provider_info["scopes_supported"] = list(set(_scopes))

        # Add claims
        _claims: List[str] = []
        for _cl in SCOPE2CLAIMS.values():
            _claims.extend(_cl)
        if self.extra_claims is not None:
            _claims.extend(self.extra_claims)
        if self.extra_scope_dict is not None:
            for _ex_claim in self.extra_scope_dict.values():
                _claims.extend(_ex_claim)
        # Remove duplicates if any
        _provider_info["claims_supported"] = list(set(_claims))

        # Sort order RS, ES, HS, PS
        sign_algs = list(jws.SIGNER_ALGS.keys())
        sign_algs = sorted(sign_algs, key=cmp_to_key(sort_sign_alg))

        # Add signing alg values
        for typ in ["userinfo", "id_token", "request_object"]:
            _provider_info["%s_signing_alg_values_supported" % typ] = sign_algs

        # Add encryption alg values
        algs = jwe.SUPPORTED["alg"]
        for typ in ["userinfo", "id_token", "request_object"]:
            _provider_info["%s_encryption_alg_values_supported" % typ] = algs

        # Add encryption enc values
        encs = jwe.SUPPORTED["enc"]
        for typ in ["userinfo", "id_token", "request_object"]:
            _provider_info["%s_encryption_enc_values_supported" % typ] = encs

        # Add acr_values
        if self.authn_broker:
            acr_values = self.authn_broker.getAcrValuesString()
            if acr_values is not None:
                _provider_info["acr_values_supported"] = acr_values

        return _provider_info

    def discovery_endpoint(self, request, handle=None, **kwargs):
        _log_debug = logger.debug

        _log_debug("@discovery_endpoint")

        request = self.server.message_factory.get_request_type(
            "discovery_endpoint"
        )().deserialize(request, "urlencoded")
        _log_debug("discovery_request:%s" % (sanitize(request.to_dict()),))

        if request["service"] != SWD_ISSUER:
            return BadRequest("Unsupported service")

        # verify that the principal is one of mine

        _response = self.server.message_factory.get_response_type("discovery_endpoint")(
            locations=[self.baseurl]
        )

        _log_debug("discovery_response:%s" % (sanitize(_response.to_dict()),))

        headers = [("Cache-Control", "no-store")]
        (key, timestamp) = handle
        if key.startswith(STR) and key.endswith(STR):
            cookie = self.cookie_func(key, self.cookie_name, "disc", self.sso_ttl)
            headers.append(cookie)

        return Response(
            _response.to_json(), content="application/json", headers=headers
        )

    def aresp_check(self, aresp, areq):
        # Use of the nonce is REQUIRED for all requests where an ID Token is
        # returned directly from the Authorization Endpoint
        if "id_token" in aresp and "nonce" not in areq:
            return error_response("invalid_request", "Missing nonce value")
        return None

    def response_mode(self, areq, fragment_enc, **kwargs):
        resp = super().response_mode(areq, fragment_enc, **kwargs)

        if resp is None and areq["response_mode"] == "form_post":
            context = {
                "action": kwargs["redirect_uri"],
                "inputs": kwargs["aresp"].to_dict(),
            }
            return Response(
                self.template_renderer("form_post", context), headers=kwargs["headers"]
            )
        return None

    def create_authn_response(self, areq, sid):
        # create the response
        aresp = self.server.message_factory.get_response_type(
            "authorization_endpoint"
        )()
        try:
            aresp["state"] = areq["state"]
        except KeyError:
            pass

        if "response_type" in areq and areq["response_type"] == ["none"]:
            fragment_enc = False
        else:
            _sinfo = self.sdb[sid]

            try:
                aresp["scope"] = areq["scope"]
            except KeyError:
                pass

            rtype = set(areq["response_type"][:])
            if len(rtype) == 1 and "code" in rtype:
                fragment_enc = False
            else:
                fragment_enc = True

            if "code" in areq["response_type"]:
                _code = aresp["code"] = self.sdb[sid]["code"]
                rtype.remove("code")
            else:
                self.sdb.update(sid, "code", None)
                _code = None

            if "token" in rtype:
                _dic = self.sdb.upgrade_to_token(issue_refresh=False, key=sid)

                logger.debug("_dic: %s" % sanitize(_dic))
                for key, val in _dic.items():
                    if key in aresp.parameters() and val is not None:
                        aresp[key] = val

                rtype.remove("token")

            try:
                _access_token = aresp["access_token"]
            except KeyError:
                _access_token = None

            if "id_token" in areq["response_type"]:
                user_info = self.userinfo_in_id_token_claims(_sinfo)
                if areq["response_type"] == ["id_token"]:
                    #  scopes should be returned here
                    info = self._collect_user_info(_sinfo)
                    if user_info is None:
                        user_info = info
                    else:
                        user_info.update(info)

                client_info = self.cdb[str(areq["client_id"])]

                hargs: Dict[str, str] = {}
                rt_set = set(areq["response_type"])
                if {"code", "id_token", "token"}.issubset(rt_set):
                    hargs = {"code": _code, "access_token": _access_token}
                elif {"code", "id_token"}.issubset(rt_set):
                    hargs = {"code": _code}
                elif {"id_token", "token"}.issubset(rt_set):
                    hargs = {"access_token": _access_token}

                # or 'code id_token'
                try:
                    id_token = self.sign_encrypt_id_token(
                        _sinfo, client_info, areq, user_info=user_info, **hargs
                    )
                except (JWEException, NoSuitableSigningKeys) as err:
                    logger.warning(str(err))
                    return error_response(
                        "invalid_request", descr="Could not sign/encrypt id_token"
                    )

                aresp["id_token"] = id_token
                _sinfo["id_token"] = id_token
                rtype.remove("id_token")

            if len(rtype):
                raise UnSupported("unsupported_response_type", list(rtype))

        return aresp, fragment_enc

    def key_setup(self, local_path, vault="keys", sig=None, enc=None):
        """
        Prepare keys for presentation.

        :param local_path: The path to where the JWKs should be stored
        :param vault: Where the private key will be stored
        :param sig: Key for signature
        :param enc: Key for encryption
        :return: A URL the RP can use to download the key.
        """
        self.jwks_uri = key_export(
            self.baseurl,
            local_path,
            vault,
            self.keyjar,
            fqdn=self.hostname,
            sig=sig,
            enc=enc,
        )

    def endsession_endpoint(self, request="", **kwargs):
        """
        Endpoint for EndSession.

        :param request:
        :param kwargs:
        :return: Either a Response instance or a tuple (Response, args)
        """
        return self.end_session_endpoint(request, **kwargs)

    def do_key_rollover(self, jwks, kid_template):
        """
        Handle key roll-over.

        Import new keys and inactivating the ones in the keyjar that are of the same type and usage.

        :param jwks: A JWKS
        :param kid_template: Key ID template
        """
        kb = KeyBundle()
        kb.do_keys(jwks["keys"])

        for k in kb.keys():
            if not k.kid:
                k.add_kid()
            self.kid[k.use][k.kty] = k.kid

            # find the old key for this key type and usage and mark that
            # as inactive
            for _kb in self.keyjar.issuer_keys[""]:
                for key in _kb.keys():
                    if key.kty == k.kty and key.use == k.use:
                        if k.kty == "EC":
                            if key.crv == k.crv:
                                key.inactive_since = time.time()
                        else:
                            key.inactive_since = time.time()

        self.keyjar.add_kb("", kb)

        if self.jwks_name:
            # print to the jwks file
            dump_jwks(self.keyjar[""], self.jwks_name)

    def remove_inactive_keys(self, more_then: int = 3600):
        """
        Remove all keys that has been inactive 'more_then' seconds.

        :param more_then: An integer (default = 3600 seconds == 1 hour)
        """
        now = time.time()
        for kb in self.keyjar.issuer_keys[""]:
            for key in kb.keys():
                if key.inactive_since:
                    if now - key.inactive_since > more_then:
                        kb.remove(key)
            if len(kb) == 0:
                self.keyjar.issuer_keys[""].remove(kb)

    def get_by_sub_and_(self, sub: str, key: str, val: Any) -> Optional[str]:
        """
        Get a session ID based on subject ID and an attribute value pair.

        Matches sessions based on a subject identifier (sub) and
        one other claim (key) having value (val).

        :param sub: The subject identifier
        :param key: A claim in the session information
        :param val: A value
        :return: A session ID
        """
        for sid in self.sdb.get_by_sub(sub):
            try:
                if self.sdb[sid][key] == val:
                    return sid
            except KeyError:
                continue
        return None

    # Below are LOGOUT related methods

    def verify_post_logout_redirect_uri(
        self, esreq: Message, client_id: str
    ) -> Optional[str]:
        """
        Verify a post logout URI.

        :param esreq: End session request
        :param client_id: The Client ID
        :return: The post logout URI if it was OK otherwise None
        """
        try:
            redirect_uri = esreq["post_logout_redirect_uri"]
        except KeyError:
            logger.debug("Missing post_logout_redirect_uri parameter")
            return None

        try:
            accepted_urls = self.cdb[client_id]["post_logout_redirect_uris"]
            if self._verify_url(redirect_uri, accepted_urls):
                return redirect_uri
        except Exception as exc:
            msg = "An error occurred while verifying redirect URI: %s"
            logger.debug(msg, str(exc))

        return None

    def let_user_verify_logout(
        self,
        uid: str,
        esr: Message,
        cookie: Optional[List[Tuple[str, str]]],
        redirect_uri: Optional[str],
    ) -> Response:
        """
        Show a page to the user, that asks whether logout should be performed.

        :param uid: User ID
        :param esr: EndSessionRequest instance
        :param cookie: A cookie
        :param redirect_uri: URL
        :return: Response instance
        """
        if cookie:
            headers = cookie
        else:
            headers = []

        self.sdb.set_verify_logout(uid)

        if redirect_uri is not None:
            redirect = redirect_uri
        else:
            redirect = "/"
        try:
            tmp_id_token_hint = esr["id_token_hint"]
        except KeyError:
            tmp_id_token_hint = ""  # nosec

        context = {
            "id_token_hint": tmp_id_token_hint,
            "post_logout_redirect_uri": esr["post_logout_redirect_uri"],
            "key": self.sdb.get_verify_logout(uid),
            "redirect": redirect,
            "action": "/" + EndSessionEndpoint("").etype,
        }
        return Response(
            self.template_renderer("verify_logout", context), headers=headers
        )

    def _get_uid_from_cookie(
        self, cookie: Optional[Union[str, SimpleCookie]]
    ) -> Tuple[Optional[CookieDealer], Optional[str], Optional[str]]:
        """
        Get cookie_dealer, client_id and uid from cookie.

        :param cookie: Received cookie
        :return: Tuple containing CookieDealer instance, client ID and User ID
        """
        if cookie is None:
            return None, None, None

        cookie_dealer = CookieDealer(srv=self)
        client_id = uid = None

        _cval = cookie_dealer.get_cookie_value(cookie, self.sso_cookie_name)
        if _cval:
            (value, _ts, typ) = _cval
            if typ == "sso":
                uid, client_id = value.split(DELIM)

        return cookie_dealer, client_id, uid

    def do_back_channel_logout(
        self, cinfo: dict, sub: str, sid: str
    ) -> Optional[Tuple[str, str]]:
        """
        Prepare information to be used to do a back-channel logout.

        :param cinfo: Client information
        :param sub: Subject identifier
        :param sid: The Issuer ID
        :return: Tuple with logout URI and signed logout token
        """
        try:
            back_channel_logout_uri = cinfo["backchannel_logout_uri"]
        except KeyError:
            return None

        # always include sub and sid so I don't check for
        # backchannel_logout_session_required

        payload = {
            "sub": sub,
            "sid": sid,
            "events": {BACK_CHANNEL_LOGOUT_EVENT: {}},
            "jti": uuid.uuid4().hex,
        }

        try:
            alg = cinfo["id_token_signed_response_alg"]
        except KeyError:
            alg = self.capabilities["id_token_signing_alg_values_supported"][0]

        _jws = JWT(self.keyjar, iss=self.name, lifetime=86400, sign_alg=alg)
        sjwt = _jws.pack(aud=cinfo["client_id"], **payload)

        return back_channel_logout_uri, sjwt

    def clean_sessions(self, usids: List[str]):
        """
        Remove Session IDs from the session DB.

        :param usids: List of session IDs
        """
        _sdb = self.sdb
        # Clean out all sessions
        for sid in usids:
            del _sdb[sid]

    def logout_info_for_all_clients(
        self, uid: Optional[str] = "", sid: Optional[str] = ""
    ) -> Dict:
        """
        Collect information necessary to logout one user from all clients he/she has been using.

        One of uid and sid MUST be provided. If uid is provided sid is ignored.
        NO changes are made to the session DB.
        No logout is actually performed
        :param uid: User ID
        :param sid: Session ID
        :return: Dictionary with logout information
        """
        if not uid:
            if not sid:
                raise ParameterError("One of uid and sid MUST be provided")
            else:
                uid = self.sdb.get_uid_by_sid(sid)

        # Find all the session IDs this user has gotten
        usids = session_get(self.sdb, "uid", uid)
        # Find all RPs this user has logged it from
        _client_sid = {}
        for usid in usids:
            _client_sid[self.sdb[usid]["client_id"]] = usid

        # Front-/Backchannel logout ?
        _cdb = self.cdb
        _iss = self.name
        bc_logouts = {}
        fc_iframes = {}
        for _cid, _csid in _client_sid.items():
            if "backchannel_logout_uri" in _cdb[_cid]:
                _sub = self.sdb[_csid]["sub"]
                bc_logouts[_cid] = self.do_back_channel_logout(_cdb[_cid], _sub, _csid)
            if "frontchannel_logout_uri" in _cdb[_cid]:
                # Construct an IFrame
                fc_iframes[_cid] = self.do_front_channel_logout_iframe(
                    _cdb[_cid], _iss, _csid
                )

        return {"back_channel": bc_logouts, "front_channel": fc_iframes}

    def logout_info_for_one_client(self, session_id: str, client_id: str) -> Dict:
        """
        Collect information necessary to log out from client.

        Note that if a client has both back channel and front channel logout registered both
        will be handled.
        :param session_id: Session ID
        :param client_id: Client ID
        :return: Dictionary with back_channel and front_channel logout info.
        """
        logout_spec: Dict[str, Dict[str, Union[None, str, Tuple[str, str]]]] = {
            "back_channel": {},  # back-channel logout information
            "front_channel": {},  # front-channel logout information
        }

        if "backchannel_logout_uri" in self.cdb[client_id]:
            _subject_id = self.sdb[session_id]["sub"]
            logout_spec["back_channel"] = {
                client_id: self.do_back_channel_logout(
                    self.cdb[client_id], _subject_id, session_id
                )
            }
        elif "frontchannel_logout_uri" in self.cdb[client_id]:
            # Construct an IFrame
            _iframe = self.do_front_channel_logout_iframe(
                self.cdb[client_id], self.name, session_id
            )
            logout_spec["front_channel"] = {client_id: _iframe}

        return logout_spec

    def end_session_endpoint(
        self,
        request: str = "",
        cookie: Optional[Union[str, SimpleCookie]] = None,
        **kwargs,
    ) -> Response:
        """
        Handle a RP initiated Logout request.

        :param request: The logout request
        :param cookie:
        :param kwargs:
        :return: Returns a dictionary with one key 'sjwt' and the value
            being a signed JWT token with session information.
        """
        _req = self.server.message_factory.get_request_type("endsession_endpoint")
        esr = _req().from_urlencoded(request)

        logger.debug("End session request: %s", sanitize(esr.to_dict()))

        if self.events:
            self.events.store("protocol request", esr)

        # 2 ways of find out client ID and user. Either through a cookie
        # or using the id_token_hint. If I get information from both make sure they match
        _, client_id, uid = self._get_uid_from_cookie(cookie)

        if uid is not None:
            client_ids = self.sdb.get_client_ids_for_uid(uid)
            if client_id not in client_ids:
                return error_response("invalid_request", "Wrong user")

        sid = ""

        if "id_token_hint" in esr:
            id_token_hint = IdToken().from_jwt(
                esr["id_token_hint"], keyjar=self.keyjar, verify=True
            )
            far_away = 86400 * 30  # 30 days

            if client_id:
                args = {"client_id": client_id}
            else:
                args = {}

            try:
                id_token_hint.verify(
                    iss=self.baseurl, skew=far_away, nonce_storage_time=far_away, **args
                )
            except (VerificationError, NotForMe) as err:
                logger.warning("Verification error on id_token_hint: %s", err)
                return error_response("invalid_request", "Bad Id Token hint")

            sub = id_token_hint["sub"]

            if uid is not None:
                # verify that 'sub' are bound to 'uid'
                if self.sdb.get_uid_by_sub(sub) != uid:
                    return error_response("invalid_request", "Wrong user")
            else:
                uid = self.sdb.get_uid_by_sub(sub)

            if client_id is None:
                if len(id_token_hint["aud"]) == 1:
                    client_id = id_token_hint["aud"][0]
                else:
                    client_id = id_token_hint["azp"]

            sids = session_get(self.sdb, "sub", sub)

            matching_client_id = False
            for sid in sids:
                if self.sdb[sid]["client_id"] == client_id:
                    matching_client_id = True
                    break

            if not matching_client_id:
                return error_response(
                    "invalid_request", "Could not find a matching client ID"
                )

        if not client_id:
            return error_response("invalid_request", "Could not find client ID")
        if client_id not in self.cdb:
            return error_response("invalid_request", "Unknown client")

        if "post_logout_redirect_uri" in esr:
            redirect_uri = self.verify_post_logout_redirect_uri(esr, client_id)
            if not redirect_uri:
                msg = "Post logout redirect URI verification failed!"
                return error_response("invalid_request", msg)
        else:  # If only one registered use that one
            try:
                _ruri = self.cdb[client_id]["post_logout_redirect_uris"]
            except KeyError:
                if self.post_logout_page is None:
                    logger.warning("No post logout page configured for %s", client_id)
                    return error_response(
                        "server_error", "Have no post logout page configured"
                    )
                else:
                    redirect_uri = self.post_logout_page
            else:
                if len(_ruri) == 1:
                    _base, _query = _ruri[0]
                    if _query:
                        query_string = urlencode(
                            [(key, v) for key in _query for v in _query[key]]
                        )
                        redirect_uri = "%s?%s" % (_base, query_string)
                    else:
                        redirect_uri = _base
                else:
                    return error_response(
                        "invalid_request",
                        descr="Missing post_logout_redirect_uri and more then one post_logout_redirect_uris",
                    )

        # redirect user to OP logout verification page
        payload = {
            "uid": uid,
            "client_id": client_id,
            "redirect_uri": redirect_uri,
            "sid": sid,
        }
        if "state" in esr:
            payload["state"] = esr["state"]

        if self.events:
            self.events.store("object args", "{}".format(payload))

        # From me to me
        _jws = JWT(
            self.keyjar, iss=self.name, lifetime=86400, sign_alg=self.signing_alg
        )
        sjwt = _jws.pack(aud=[self.name], **payload)

        location = "{}?{}".format(self.logout_verify_url, urlencode({"sjwt": sjwt}))
        return SeeOther(location)

    def unpack_signed_jwt(self, sjwt: str):
        """Will unpack a signed JWT."""
        verifier = JWT(self.keyjar)
        try:
            return verifier.unpack(sjwt)
        except Exception as err:
            raise ValueError(err)

    def do_verified_logout(
        self, sid: str, client_id: str, alla: bool = False, **kwargs
    ) -> Union[dict, Dict[str, list]]:
        """
        Perform back channel logout and prepares the information needed for front channel logout.

        :param sid: Session ID
        :param client_id: Client ID
        :param alla: Whether logout should be attempted from all clients or just one specific client.
        :param kwargs:
        :return:
        """
        if alla:
            uid = self.sdb.get_uid_by_sid(sid)
            logout_spec = self.logout_info_for_all_clients(uid)
            # Find all the session IDs this user has gotten
            sids = session_get(self.sdb, "uid", uid)
        else:
            logout_spec = self.logout_info_for_one_client(
                session_id=sid, client_id=client_id
            )
            sids = [sid]

        if self.events:
            self.events.store("object args", "{}".format(logout_spec))

        if not logout_spec["back_channel"] and not logout_spec["front_channel"]:
            # kill cookies
            kaka1 = self.write_session_cookie(
                "removed", http_only=False, same_site="None"
            )
            kaka2 = self.cookie_func(
                "", typ="sso", cookie_name=self.sso_cookie_name, kill=True
            )
            return {"cookie": [kaka1, kaka2]}

        # take care of Back channel logout first
        if logout_spec["back_channel"]:
            failed = []
            headers = {"Content-Type": "application/x-www-form-urlencoded"}
            for _cid, spec in logout_spec["back_channel"].items():
                _url, sjwt = spec
                logger.info("logging out from {} at {}".format(_cid, _url))

                try:
                    resp = self.httpc.http_request(
                        _url,
                        "POST",
                        data="logout_token={}".format(sjwt),
                        headers=headers,
                    )
                except Exception as err:
                    # Can't be more specific because I don't know which http client are used
                    logger.error("failed to logout from {}".format(_cid))
                    if self.events:
                        self.events.store("exception", "{}: {}".format(_cid, str(err)))
                    failed.append(_cid)
                    continue

                if resp.status_code < 300:
                    logger.info("Logged out from {}".format(_cid))
                else:
                    _errstr = "failed to logout from {}".format(_cid)
                    if self.events:
                        self.events.store("fault", _errstr)
                    logger.error(_errstr)
                    failed.append(_cid)
            # If no back-channel logout worked and there is no front-channel logout
            # regard this as a failure.
            if len(failed) == len(logout_spec["back_channel"]):
                if not logout_spec["front_channel"]:
                    return {}

        # kill cookies
        kaka1 = self.write_session_cookie("removed", http_only=False, same_site="None")
        kaka2 = self.cookie_func(
            "", typ="sso", cookie_name=self.sso_cookie_name, kill=True
        )
        res = {"cookie": [kaka1, kaka2]}

        if logout_spec["front_channel"]:
            for _cid in logout_spec["front_channel"].keys():
                logger.info("Adding logout iframe for {}".format(_cid))
            res["iframe"] = list(logout_spec["front_channel"].values())

        # Clean out all sessions
        self.clean_sessions(sids)

        return res

    @staticmethod
    def do_front_channel_logout_iframe(
        client_info: Dict, issuer: str, session_id: str
    ) -> Optional[str]:
        """
        Construct a front channel logout IFrame.

        :param client_info: Client info
        :param issuer: Issuer ID
        :param session_id: Session ID
        :return: HTML IFrame string
        """
        try:
            frontchannel_logout_uri = client_info["frontchannel_logout_uri"]
        except KeyError:
            return None

        try:
            flsr = client_info["frontchannel_logout_session_required"]
        except KeyError:
            flsr = False

        if flsr:
            _query = {"iss": issuer, "sid": session_id}
            if "?" in frontchannel_logout_uri:
                p = urlparse(frontchannel_logout_uri)
                _args = {k: v[0] for k, v in parse_qs(p.query).items()}
                _args.update(_query)
                _query = _args
                _np = p._replace(query="")
                frontchannel_logout_uri = _np.geturl()

            _iframe = '<iframe src="{}?{}">'.format(
                frontchannel_logout_uri, urlencode(_query)
            )
        else:
            _iframe = '<iframe src="{}">'.format(frontchannel_logout_uri)

        return _iframe



import logging
import warnings
from typing import Any
from typing import Dict
from typing import Optional

from oic import rndstr
from oic.oauth2.message import REQUIRED_LIST_OF_STRINGS
from oic.oauth2.message import SINGLE_OPTIONAL_STRING
from oic.oauth2.message import SINGLE_REQUIRED_STRING
from oic.oauth2.message import Message
from oic.oic import REQUEST2ENDPOINT
from oic.oic import RESPONSE2ERROR
from oic.oic import Client
from oic.oic import Server as OicServer
from oic.oic.message import Claims
from oic.oic.message import OpenIDSchema
from oic.oic.provider import Endpoint
from oic.oic.provider import Provider
from oic.utils.http_util import Response
from oic.utils.keyio import KeyJar
from oic.utils.sanitize import sanitize
from oic.utils.settings import OicClientSettings
from oic.utils.settings import OicProviderSettings

__author__ = "rohe0002"

logger = logging.getLogger(__name__)


class UserClaimsRequest(Message):
    c_param = {
        "sub": SINGLE_REQUIRED_STRING,
        "client_id": SINGLE_REQUIRED_STRING,
        "client_secret": SINGLE_REQUIRED_STRING,
        "claims_names": REQUIRED_LIST_OF_STRINGS,
    }


class UserClaimsResponse(Message):
    c_param = {
        "claims_names": REQUIRED_LIST_OF_STRINGS,
        "jwt": SINGLE_OPTIONAL_STRING,
        "endpoint": SINGLE_OPTIONAL_STRING,
        "access_token": SINGLE_OPTIONAL_STRING,
    }


class UserInfoClaimsRequest(Message):
    c_param = {"access_token": SINGLE_REQUIRED_STRING}


class OICCServer(OicServer):
    def parse_user_claims_request(self, info, sformat="urlencoded"):
        return self._parse_request(UserClaimsRequest, info, sformat)

    def parse_userinfo_claims_request(self, info, sformat="urlencoded"):
        return self._parse_request(UserInfoClaimsRequest, info, sformat)


class ClaimsServer(Provider):
    def __init__(
        self,
        name,
        sdb,
        cdb,
        userinfo,
        client_authn,
        urlmap=None,
        keyjar=None,
        hostname="",
        dist_claims_mode=None,
        verify_ssl=None,
        settings=None,
    ):
        self.settings = settings or OicProviderSettings()
        if verify_ssl is not None:
            warnings.warn(
                "`verify_ssl` is deprecated, please use `settings` instead if you need to set a non-default value.",
                DeprecationWarning,
                stacklevel=2,
            )
            self.settings.verify_ssl = verify_ssl
        Provider.__init__(
            self,
            name,
            sdb,
            cdb,
            None,
            userinfo,
            None,
            client_authn,
            None,
            urlmap,
            keyjar,
            hostname,
            settings=self.settings,
        )

        if keyjar is None:
            keyjar = KeyJar(verify_ssl=verify_ssl)

        for cid, _dic in cdb.items():
            try:
                keyjar.add_symmetric(cid, _dic["client_secret"], ["sig", "ver"])
            except KeyError:
                pass

        self.srvmethod = OICCServer(keyjar=keyjar)
        self.dist_claims_mode = dist_claims_mode
        self.info_store: Dict[str, Any] = {}
        self.claims_userinfo_endpoint = ""

    def _aggregation(self, info):

        jwt_key = self.keyjar.get_signing_key()
        cresp = UserClaimsResponse(
            jwt=info.to_jwt(key=jwt_key, algorithm="RS256"),
            claims_names=list(info.keys()),
        )

        logger.info("RESPONSE: %s" % (sanitize(cresp.to_dict()),))
        return cresp

    def _distributed(self, info):
        # store the user info so it can be accessed later
        access_token = rndstr()
        self.info_store[access_token] = info
        return UserClaimsResponse(
            endpoint=self.claims_userinfo_endpoint,
            access_token=access_token,
            claims_names=info.keys(),
        )

    def do_aggregation(self, info, uid):
        return self.dist_claims_mode.aggregate(uid, info)

    def claims_endpoint(self, request, http_authz, *args):
        _log_info = logger.info

        ucreq = self.srvmethod.parse_user_claims_request(request)

        _log_info("request: %s" % sanitize(ucreq))

        try:
            self.client_authn(self, ucreq, http_authz)
        except Exception as err:
            _log_info("Failed to verify client due to: %s" % err)

        if "claims_names" in ucreq:
            claim_args = dict([(n, {"optional": True}) for n in ucreq["claims_names"]])
            uic: Optional[Claims] = Claims(**claim_args)
        else:
            uic = None

        _log_info("User info claims: %s" % sanitize(uic))

        # oicsrv, userdb, subject, client_id="", user_info_claims=None
        info = self.userinfo(
            ucreq["sub"], user_info_claims=uic, client_id=ucreq["client_id"]
        )

        _log_info("User info: %s" % sanitize(info))

        # Convert to message format
        info = OpenIDSchema(**info)

        if self.do_aggregation(info, ucreq["sub"]):
            cresp = self._aggregation(info)
        else:
            cresp = self._distributed(info)

        _log_info("response: %s" % sanitize(cresp.to_dict()))

        return Response(cresp.to_json(), content="application/json")

    def claims_info_endpoint(self, request, authn):
        _log_info = logger.info

        _log_info("Claims_info_endpoint query: '%s'" % sanitize(request))

        ucreq = self.srvmethod.parse_userinfo_claims_request(request)
        # Access_token is mandatory in UserInfoClaimsRequest
        uiresp = OpenIDSchema(**self.info_store[ucreq["access_token"]])

        _log_info("returning: %s" % sanitize(uiresp.to_dict()))
        return Response(uiresp.to_json(), content="application/json")


class ClaimsClient(Client):
    def __init__(self, client_id=None, verify_ssl=None, settings=None):
        self.settings = settings or OicClientSettings()
        if verify_ssl is not None:
            warnings.warn(
                "`verify_ssl` is deprecated, please use `settings` instead if you need to set a non-default value.",
                DeprecationWarning,
                stacklevel=2,
            )
            self.settings.verify_ssl = verify_ssl

        Client.__init__(self, client_id, settings=self.settings)

        self.request2endpoint = REQUEST2ENDPOINT.copy()
        self.request2endpoint["UserClaimsRequest"] = "userclaims_endpoint"
        self.response2error = RESPONSE2ERROR.copy()
        self.response2error["UserClaimsResponse"] = ["ErrorResponse"]

    def construct_UserClaimsRequest(
        self, request=UserClaimsRequest, request_args=None, extra_args=None, **kwargs
    ):

        return self.construct_request(request, request_args, extra_args)

    def do_claims_request(
        self,
        request=UserClaimsRequest,
        request_resp=UserClaimsResponse,
        body_type="json",
        method="POST",
        request_args=None,
        extra_args=None,
        http_args=None,
    ):

        url, body, ht_args, _ = self.request_info(
            request, method=method, request_args=request_args, extra_args=extra_args
        )

        if http_args is None:
            http_args = ht_args
        else:
            http_args.update(http_args)

        # http_args = self.init_authentication_method(csi, "bearer_header",
        #                                                    request_args)

        return self.request_and_return(
            url,
            request_resp,
            method,
            body,
            body_type,
            extended=False,
            http_args=http_args,
            key=self.keyjar.verify_keys(self.keyjar.match_owner(url)),
        )


class UserClaimsEndpoint(Endpoint):
    etype = "userclaims"


class UserClaimsInfoEndpoint(Endpoint):
    etype = "userclaimsinfo"



from typing import Dict

from oic import rndstr
from oic.extension.single import SingleClient
from oic.extension.single import SingleService
from oic.oauth2.message import SINGLE_OPTIONAL_INT
from oic.oauth2.message import SINGLE_OPTIONAL_STRING
from oic.oauth2.message import SINGLE_REQUIRED_STRING
from oic.oauth2.message import AccessTokenResponse
from oic.oauth2.message import Message
from oic.utils.time_util import time_sans_frac


class AuthorizationRequest(Message):
    c_param = {
        "response_type": SINGLE_REQUIRED_STRING,
        "client_id": SINGLE_REQUIRED_STRING,
        "scope": SINGLE_OPTIONAL_STRING,
    }


class AuthorizationResponse(Message):
    c_param = {
        "device_code": SINGLE_REQUIRED_STRING,
        "user_code": SINGLE_REQUIRED_STRING,
        "verification_uri": SINGLE_REQUIRED_STRING,
        "expires_in": SINGLE_OPTIONAL_INT,
        "interval": SINGLE_OPTIONAL_INT,
    }


class TokenRequest(Message):
    c_param = {
        "grant_type": SINGLE_REQUIRED_STRING,
        "device_code": SINGLE_REQUIRED_STRING,
        "client_id": SINGLE_REQUIRED_STRING,
    }


class DeviceFlowServer(SingleService):
    def __init__(self, host):
        SingleService.__init__(self, host)
        self.host = host
        # map between device_code and user_code
        self.client_id2device: Dict[str, str] = {}
        self.device2user: Dict[str, str] = {}
        self.user_auth: Dict[str, bool] = {}
        self.device_code_expire_at: Dict[str, int] = {}
        self.device_code_life_time = 900  # 15 minutes

    def device_endpoint(self, request, authn=None):

        _req = AuthorizationRequest(**request)
        device_code = rndstr(10)
        user_code = rndstr(6)

        self.device2user[device_code] = user_code
        self.user_auth[user_code] = False
        self.client_id2device[_req["client_id"]] = device_code
        self.device_code_expire_at[device_code] = (
            time_sans_frac() + self.device_code_life_time
        )

    def token_endpoint(self, request, authn=None):
        _req = TokenRequest(**request)
        _dc = _req["device_code"]

        if time_sans_frac() > self.device_code_expire_at[_dc]:
            return self.host.error_code(error="expired_token")

        _uc = self.device2user[_dc]

        if self.user_auth[_uc]:  # User is authenticated
            pass
        else:
            return self.host.error_code(error="authorization_pending")

    def device_auth(self, user_code):
        self.user_auth[user_code] = True


class DeviceFlowClient(SingleClient):
    def __init__(self, host):
        SingleClient.__init__(self, host)
        self.requests = {
            "authorization": self.authorization_request,
            "token": self.authorization_request,
        }

    def authorization_request(self, scope=""):
        req = AuthorizationRequest(
            client_id=self.host.client_id, response_type="device_code"
        )
        if scope:
            req["scope"] = scope

        http_response = self.host.http_request(
            self.host.provider_info["device_endpoint"], "POST", req.to_urlencoded()
        )

        response = self.host.parse_request_response(
            AuthorizationResponse, http_response, "json"
        )

        return response

    def token_request(self, device_code=""):
        req = TokenRequest(
            grant_type="urn:ietf:params:oauth:grant-type:device_code",
            device_code=device_code,
            client_id=self.host.client_id,
        )

        http_response = self.host.http_request(
            self.host.provider_info["token_endpoint"], "POST", req.to_urlencoded()
        )

        response = self.host.parse_request_response(
            AccessTokenResponse, http_response, "json"
        )

        return response



from urllib.parse import urlparse

from oic.oauth2.message import REQUIRED_LIST_OF_STRINGS
from oic.oauth2.message import SINGLE_REQUIRED_STRING
from oic.oic.message import SINGLE_REQUIRED_INT
from oic.oic.message import JasonWebToken
from oic.utils.keyio import KeyBundle

__author__ = "roland"


class PrivateKeyJWT(JasonWebToken):
    c_param = JasonWebToken.c_param.copy()
    c_param.update(
        {
            "aud": SINGLE_REQUIRED_STRING,
            "iss": SINGLE_REQUIRED_STRING,
            "sub": SINGLE_REQUIRED_STRING,
            "aud": SINGLE_REQUIRED_STRING,
            "exp": SINGLE_REQUIRED_INT,
            "iat": SINGLE_REQUIRED_INT,
            "jti": SINGLE_REQUIRED_STRING,
        }
    )


def verify_url(url):
    """
    Verify security of URL.

    Hosted on a website with Transport Layer Security (TLS) protection
    (a Hypertext Transfer Protocol – Secure (HTTPS) URI)
    Hosted on the local domain of the client (e.g., http://localhost/)
    Hosted on a client-specific non-remote-protocol URI scheme (e.g., myapp://)

    :param url:
    :return:
    """
    if url.startswith("http://localhost"):
        return True
    else:
        p = urlparse(url)
        if p.scheme == "http":
            return False

    return True


class HeartSoftwareStatement(JasonWebToken):
    c_param = JasonWebToken.c_param.copy()
    c_param.update(
        {
            "redirect_uris": REQUIRED_LIST_OF_STRINGS,
            "grant_types": SINGLE_REQUIRED_STRING,
            "jwks_uri": SINGLE_REQUIRED_STRING,
            "jwks": SINGLE_REQUIRED_STRING,
            "client_name": SINGLE_REQUIRED_STRING,
            "client_uri": SINGLE_REQUIRED_STRING,
        }
    )
    c_allowed_values = {"grant_types": ["authorization_code", "implicit"]}

    def verify(self, **kwargs):
        if "jwks" in self:
            try:
                _keys = self["jwks"]["keys"]
            except KeyError:
                raise SyntaxError('"keys" parameter missing')
            else:
                # will raise an exception if syntax error
                KeyBundle(_keys)
        for param in ["jwks_uri", "client_uri"]:
            verify_url(self[param])

        JasonWebToken.verify(self, **kwargs)



import shelve  # nosec

__author__ = "danielevertsson"


class ShelfWrapper(object):
    def __init__(self, filename):
        self.filename = filename

    def keys(self):
        db = self._reopen_database()
        return db.keys()

    def __len__(self):
        db = self._reopen_database()
        return db.__len__()

    def has_key(self, key):
        return key in self

    def __contains__(self, key):
        db = self._reopen_database()
        return db.__contains__(key)

    def get(self, key, default=None):
        db = self._reopen_database()
        return db.get(key, default)

    def __getitem__(self, key):
        db = self._reopen_database()
        return db.__getitem__(key)

    def __setitem__(self, key, value):
        db = self._reopen_database()
        db.__setitem__(key, value)

    def __delitem__(self, key):
        db = self._reopen_database()
        db.__delitem__(key)

    def _reopen_database(self):
        return shelve.open(self.filename, writeback=True)  # nosec


def open(filename):
    """
    Open a persistent dictionary for reading and writing.

    The filename parameter is the base filename for the underlying
    database.  As a side-effect, an extension may be added to the
    filename and more than one file may be created.  The optional flag
    parameter has the same interpretation as the flag parameter of
    anydbm.open(). The optional protocol parameter specifies the
    version of the pickle protocol (0, 1, or 2).

    See the module's __doc__ string for an overview of the interface.
    """
    return ShelfWrapper(filename)



import base64
import copy
import hashlib
import json
import logging
import uuid
import warnings
from binascii import Error
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from cryptography.fernet import Fernet
from cryptography.fernet import InvalidToken

from oic import rndstr
from oic.exception import ImproperlyConfigured
from oic.oauth2.message import AuthorizationRequest
from oic.utils import tobytes
from oic.utils.session_backend import AuthnEvent
from oic.utils.session_backend import DictSessionBackend
from oic.utils.session_backend import SessionBackend
from oic.utils.time_util import utc_time_sans_frac

__author__ = "rohe0002"

logger = logging.getLogger(__name__)


def lv_pack(*args):
    s = []
    for a in args:
        s.append("{}:{}".format(len(a), a))
    return "".join(s)


def lv_unpack(txt):
    txt = txt.strip()
    res = []
    while txt:
        l, v = txt.split(":", 1)
        res.append(v[: int(l)])
        txt = v[int(l) :]
    return res


class ExpiredToken(Exception):
    pass


class WrongTokenType(Exception):
    pass


class AccessCodeUsed(Exception):
    pass


class UnknownToken(Exception):
    pass


def pairwise_id(sub, sector_identifier, seed):
    return hashlib.sha256(
        ("%s%s%s" % (sub, sector_identifier, seed)).encode("utf-8")
    ).hexdigest()


class Crypt(object):
    def __init__(self, password, mode=None):
        self.key = base64.urlsafe_b64encode(
            hashlib.sha256(password.encode("utf-8")).digest()
        )
        self.core = Fernet(self.key)

    def encrypt(self, text):
        # Padding to blocksize of AES
        text = tobytes(text)
        if len(text) % 16:
            text += b" " * (16 - len(text) % 16)
        return self.core.encrypt(tobytes(text))

    def decrypt(self, ciphertext):
        return self.core.decrypt(ciphertext)


class Token(object):
    def __init__(self, typ, lifetime=0, token_storage=None, **kwargs):
        self.type = typ
        self.lifetime = lifetime
        self.args = kwargs
        if typ == "R":
            if token_storage is None:
                raise ImproperlyConfigured(
                    "token_storage kwarg must be passed in for refresh token."
                )
        self.token_storage = token_storage

    def __call__(self, sid, *args, **kwargs):
        """
        Return a token.

        :param sid: Session id
        :return:
        """
        raise NotImplementedError()

    def key(self, **kwargs):
        """Return a key - the session id."""
        return rndstr(32)

    def type_and_key(self, token):
        """
        Return type of Token (A=Access code, T=Token, R=Refresh token) and the session id.

        :param token: A token
        :return: tuple of token type and session id
        """
        raise NotImplementedError()

    def get_key(self, token):
        """
        Return session id.

        :param token: A token
        :return: The session id
        """
        raise NotImplementedError()

    def get_type(self, token):
        """
        Return token type.

        :param token: A token
        :return: Type of Token (A=Access code, T=Token, R=Refresh token)
        """
        raise NotImplementedError()

    def expires_at(self, token):
        """
        Return the expiry timestamp of the token.

        :param token: A token
        :return: Timestamp of the token expiry in UTC
        """
        raise NotImplementedError()

    def is_expired(self, token, when=None):
        """Return if token is still valid."""
        if when is None:
            now = utc_time_sans_frac()
        else:
            now = when
        eat = self.expires_at(token)
        return bool(now > eat)

    def invalidate(self, token):
        """Mark the refresh token as invalidated."""
        if self.get_type(token) != "R":
            return
        sid = self.get_key(token)
        self.token_storage[sid]["revoked"] = True

    def valid(self, token):
        try:
            typ, key = self.type_and_key(token)
        except (Error, InvalidToken):
            raise WrongTokenType()
        if typ != self.type:
            raise WrongTokenType()
        if typ == "R":
            return not self.token_storage[key].get("revoked", False)
        else:
            return True


class DefaultToken(Token):
    def __init__(self, secret, password, typ="", **kwargs):
        Token.__init__(self, typ, **kwargs)
        self.crypt = Crypt(password)

    def __call__(self, sid="", ttype="", **kwargs):
        """
        Return a token.

        :param ttype: Type of token
        :param prev: Previous token, if there is one to go from
        :param sid: Session id
        :return:
        """
        if not ttype and self.type:
            ttype = self.type
        else:
            ttype = "A"

        tmp = ""
        rnd = ""
        while rnd == tmp:  # Don't use the same random value again
            rnd = rndstr(32)  # Ultimate length multiple of 16

        issued_at = "{}".format(utc_time_sans_frac())
        if ttype == "R":
            # kwargs["sinfo"] is a dictionary and we do not want updates...
            self.token_storage[sid] = copy.deepcopy(kwargs["sinfo"])

        return base64.b64encode(
            self.crypt.encrypt(lv_pack(rnd, ttype, sid, issued_at).encode())
        ).decode("utf-8")

    def key(self, user="", areq=None):
        """
        Return a key - the session id - that are based on some session connected data.

        :param user: User id
        :param areq: The authorization request
        :return: A hash
        """
        csum = hashlib.new("sha224")
        csum.update(rndstr(32).encode("utf-8"))
        return csum.hexdigest()  # 56 bytes long, 224 bits

    def _split_token(self, token):
        """
        Decode the token.

        :param token: A token
        :return: Tuple of sid, type, iat, salt
        """
        plain = self.crypt.decrypt(base64.b64decode(token)).decode()
        # order: rnd, type, sid, iat
        p = lv_unpack(plain)
        return p[1], p[2], int(p[3]), p[0]

    def type_and_key(self, token):
        """
        Return type of Token (A=Access code, T=Token, R=Refresh token) and the session id.

        :param token: A token
        :return: tuple of token type and session id
        """
        a, b, _, _ = self._split_token(token)
        return a, b

    def get_key(self, token):
        """
        Return session id.

        :param token: A token
        :return: The session id
        """
        return self._split_token(token)[1]

    def get_type(self, token):
        """
        Return token type.

        :param token: A token
        :return: Type of Token (A=Access code, T=Token, R=Refresh token)
        """
        return self._split_token(token)[0]

    def expires_at(self, token):
        """
        Return expiry time.

        :param token: A token
        :return: expiry timestamp
        """
        return self._split_token(token)[2] + self.lifetime


class RefreshDB(object):
    """Database for refresh token storage."""

    def __init__(self):
        warnings.warn(
            "Using `RefreshDB` is deprecated, please use `Token` and `refresh_token_factory` instead.",
            DeprecationWarning,
            stacklevel=2,
        )

    def get(self, refresh_token):
        """
        Retrieve info about the authentication proces from the refresh token.

        :return: Dictionary with info
        :raises: KeyError
        """
        raise NotImplementedError

    def store(self, token, info):
        """
        Store the information about the authentication process.

        :param token: Token
        :param info: Information associated with token to be stored
        """
        raise NotImplementedError

    def remove(self, token):
        """
        Remove the token and related information from the internal storage.

        :param token: Token to be removed
        """
        raise NotImplementedError

    def create_token(self, client_id, uid, scopes, sub, authzreq, sid):
        """
        Create refresh token for given combination of client_id and sub and store it in internal storage.

        :param client_id: Client_id of the consumer
        :param uid: User identification
        :param scopes: Scopes associated with the token
        :param sub: Sub identifier
        :param authzreq: Authorization request
        :param sid: Session ID
        :return: Refresh token
        """
        refresh_token = "Refresh_{}".format(rndstr(5 * 16))
        self.store(
            refresh_token,
            {
                "client_id": client_id,
                "uid": uid,
                "scope": scopes,
                "sub": sub,
                "authzreq": authzreq,
                "sid": sid,
            },
        )
        return refresh_token

    def verify_token(self, client_id, refresh_token):
        """Verify if the refresh token belongs to client_id."""
        if not refresh_token.startswith("Refresh_"):
            raise WrongTokenType
        try:
            stored_cid = self.get(refresh_token).get("client_id")
        except KeyError:
            return False
        return client_id == stored_cid

    def revoke_token(self, token):
        """Remove token from database."""
        self.remove(token)


class DictRefreshDB(RefreshDB):
    """Dictionary based implementation of RefreshDB."""

    def __init__(self):
        super(DictRefreshDB, self).__init__()
        warnings.warn(
            "Using `DictRefreshDB` is deprecated, please use `Token` and `refresh_token_factory` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self._db: Dict[str, Dict[str, str]] = {}

    def get(self, refresh_token):
        """Retrieve info for given token from dictionary."""
        return self._db[refresh_token].copy()

    def store(self, token, info):
        """Add token and info to the dictionary."""
        self._db[token] = info

    def remove(self, token):
        """Remove the token from the dictionary."""
        self._db.pop(token)


def create_session_db(
    base_url,
    secret,
    password,
    db=None,
    token_expires_in=3600,
    grant_expires_in=600,
    refresh_token_expires_in=86400,
):
    """
    Construct SessionDB instance.

    Using this you can create a very basic non persistent
    session database that issues opaque DefaultTokens.

    :param base_url: Same as base_url parameter of `SessionDB`.
    :param secret: Secret to pass to `DefaultToken` class.
    :param password: Secret key to pass to `DefaultToken` class.
    :param db: Storage for the session data, usually a dict.
    :param token_expires_in: Expiry time for access tokens in seconds.
    :param grant_expires_in: Expiry time for access codes in seconds.
    :param refresh_token_expires_in: Expiry time for refresh tokens.

    :return: A constructed `SessionDB` object.
    """
    code_factory = DefaultToken(secret, password, typ="A", lifetime=grant_expires_in)
    token_factory = DefaultToken(secret, password, typ="T", lifetime=token_expires_in)
    db = DictSessionBackend() if db is None else db
    refresh_token_factory = DefaultToken(
        secret, password, typ="R", lifetime=refresh_token_expires_in, token_storage={}
    )

    return SessionDB(
        base_url,
        db,
        refresh_db=None,
        code_factory=code_factory,
        token_factory=token_factory,
        refresh_token_factory=refresh_token_factory,
    )


class SessionDB(object):
    def __init__(
        self,
        base_url,
        db,
        refresh_db=None,
        refresh_token_expires_in=None,
        token_factory=None,
        code_factory=None,
        refresh_token_factory=None,
        sm_salt="",
    ):
        """
        Object to store the session related information.

        :param db: Database for storing the session information.
        """
        if refresh_token_expires_in is not None:
            warnings.warn(
                "Setting a `refresh_token_expires_in` has no effect, please set the expiration on "
                "`refresh_token_factory`.",
                DeprecationWarning,
            )
        self.base_url = base_url
        if not isinstance(db, SessionBackend):
            warnings.warn(
                "Please use `SessionBackend` to ensure proper API for the database.",
                DeprecationWarning,
            )
        self._db = db

        self.sm_salt = sm_salt or rndstr(32)

        self.token_factory = {"code": code_factory, "access_token": token_factory}

        self.token_factory_order = ["code", "access_token"]

        # TODO: This should simply be a factory like all the others too,
        #       even for the default case.

        if refresh_token_factory:
            if refresh_db:
                raise ImproperlyConfigured(
                    "Only use one of refresh_db or refresh_token_factory"
                )
            self._refresh_db = None
            self.token_factory["refresh_token"] = refresh_token_factory
            self.token_factory_order.append("refresh_token")
        elif refresh_db:
            warnings.warn(
                "Using `refresh_db` is deprecated, please use `refresh_token_factory`",
                DeprecationWarning,
                stacklevel=2,
            )
            self._refresh_db = refresh_db
        else:
            # Not configured
            self._refresh_db = None
            self.token_factory["refresh_token"] = None

        self.access_token = self.token_factory["access_token"]
        self.token = self.access_token

    def _get_token_key(self, item, order=None):
        if order is None:
            order = self.token_factory_order

        for key in order:
            try:
                return self.token_factory[key].get_key(item)
            except Exception:  # nosec
                # FIXME: Catch specific exception
                pass

        logger.info("Unknown token format")
        raise KeyError(item)

    def _get_token_type_and_key(self, item, order=None):
        if order is None:
            order = self.token_factory_order

        for key in order:
            try:
                return self.token_factory[key].type_and_key(item)
            except Exception:  # nosec
                # FIXME: Catch specific exception.
                pass

        logger.info("Unknown token format")
        raise KeyError(item)

    def _get_token_type(self, item, order=None):
        if order is None:
            order = self.token_factory_order

        for key in order:
            try:
                return self.token_factory[key].get_type(item)
            except Exception:  # nosec
                # FIXME: Catch specific exception
                pass

        logger.info("Unknown token format or invalid token")
        raise KeyError(item)

    def __getitem__(self, item):
        """
        Return Session item.

        :param item: authz grant code or refresh token
        """
        try:
            return self._db[item]
        except KeyError:
            sid = self._get_token_key(item)
            return self._db[sid]

    def __setitem__(self, key, value):
        """
        Assign Session item.

        :param key: authz grant code or refresh token
        """
        self._db[key] = value

    def __delitem__(self, sid):
        """
        Actually delete the pointed session from this SessionDB instance.

        :param sid: session identifier
        """
        del self._db[sid]

    def update(self, key, attribute, value):
        if key in self._db:
            pass
        else:
            sid = self._get_token_key(key)

            if sid not in self._db:
                raise KeyError
            else:
                key = sid

        item = self._db[key]
        item[attribute] = value
        self._db[key] = item

    def update_by_token(self, token, attribute, value):
        (typ, key) = self._get_token_type_and_key(token)
        return self.update(key, attribute, value)

    def do_sub(self, sid, client_salt, sector_id="", subject_type="public"):
        """
        Construct a sub (subject identifier).

        :param sid: Session identifier
        :param sector_id: Possible sector identifier
        :param subject_type: 'public'/'pairwise'
        :param client_salt: client specific salt - used in pairwise
        :return:
        """
        authn_event = self.get_authentication_event(sid)
        uid = authn_event.uid
        user_salt = authn_event.salt

        if subject_type == "public":
            sub = hashlib.sha256(
                "{}{}".format(uid, user_salt).encode("utf-8")
            ).hexdigest()
        else:
            sub = pairwise_id(uid, sector_id, "{}{}".format(client_salt, user_salt))

        self.update(sid, "sub", sub)

        return sub

    def create_authz_session(self, aevent, areq, id_token=None, oidreq=None, **kwargs):
        """
        Create session holding info about the Authorization event.

        :param aevent: An AuthnEvent instance
        :param areq: The AuthorizationRequest instance
        :param id_token: An IDToken instance
        :param oidreq: An OpenIDRequest instance
        :return: The session identifier, which is the database key
        """
        sid = self.token_factory["code"].key(user=aevent.uid, areq=areq)
        access_grant = self.token_factory["code"](sid=sid)

        _dic = {
            "oauth_state": "authz",
            "code": access_grant,
            "code_used": False,
            "authzreq": areq.to_json(),
            "client_id": areq["client_id"],
            "response_type": areq["response_type"],
            "revoked": False,
            "authn_event": aevent.to_json(),
        }

        _dic.update(kwargs)

        try:
            _val = areq["nonce"]
            if _val:
                _dic["nonce"] = _val
        except (AttributeError, KeyError):
            pass

        for key in ["redirect_uri", "state", "scope", "si_redirects"]:
            try:
                _dic[key] = areq[key]
            except KeyError:
                pass

        if id_token:
            _dic["id_token"] = id_token
        if oidreq:
            _dic["oidreq"] = oidreq.to_json()

        self._db[sid] = _dic

        return sid

    def get_authentication_event(self, sid):
        """Return AuthnEvent based on sid."""
        # This is a compatibility shim for older sessions
        if isinstance(self._db[sid]["authn_event"], dict):
            return AuthnEvent(**self._db[sid]["authn_event"])
        else:
            return AuthnEvent.from_json(self._db[sid]["authn_event"])

    def get_token(self, sid):
        if self._db[sid]["oauth_state"] == "authz":
            return self._db[sid]["code"]
        elif self._db[sid]["oauth_state"] == "token":
            return self._db[sid]["access_token"]

    def upgrade_to_token(  # nosec
        self,
        token=None,
        issue_refresh=False,
        id_token="",
        oidreq=None,
        key=None,
        access_grant="",
    ):
        """
        Promote session to token.

        :param token: The access grant
        :param issue_refresh: If a refresh token should be issued
        :param id_token: An IDToken instance
        :param oidreq: An OpenIDRequest instance
        :param key: The session key. One of token or key must be given.
        :return: The session information as a dictionary
        """
        if token:
            try:
                (_, key) = self.token_factory["code"].type_and_key(token)
            except Exception:
                raise WrongTokenType("Not a grant token")

            dic = self._db[key]

            if dic["code_used"]:
                raise AccessCodeUsed()
            _at = self.access_token(sid=key, sinfo=dic)
            dic["code_used"] = True
        else:
            dic = self._db[key]
            _at = self.access_token(sid=key, sinfo=dic)

        dic["access_token"] = _at
        dic["access_token_scope"] = "?"  # nosec
        dic["oauth_state"] = "token"
        dic["token_type"] = "Bearer"  # nosec

        if id_token:
            dic["id_token"] = id_token
        if oidreq:
            dic["oidreq"] = oidreq

        if issue_refresh:
            if "authn_event" in dic:
                authn_event = AuthnEvent.from_json(dic["authn_event"])
            else:
                authn_event = None
            if authn_event:
                uid = authn_event.uid
            else:
                uid = None

            if self._refresh_db:
                refresh_token = self._refresh_db.create_token(
                    dic["client_id"],
                    uid,
                    dic.get("scope"),
                    dic["sub"],
                    dic["authzreq"],
                    key,
                )
                dic["refresh_token"] = refresh_token
            elif self.token_factory["refresh_token"] is not None:
                refresh_token = self.token_factory["refresh_token"](key, sinfo=dic)
                dic["refresh_token"] = refresh_token
        self._db[key] = dic
        return dic

    def refresh_token(self, rtoken, client_id):
        """
        Issue a new access token using a valid refresh token.

        :param rtoken: Refresh token
        :param client_id: Client ID
        :return: Dictionary with session info
        :raises: ExpiredToken for invalid refresh token
                 WrongTokenType for wrong token type
        """
        # assert that it is a refresh token and that it is valid
        if self._refresh_db:
            if self._refresh_db.verify_token(client_id, rtoken):
                # Valid refresh token
                _info = self._refresh_db.get(rtoken)
                try:
                    sid = _info["sid"]
                except KeyError:
                    areq = json.loads(_info["authzreq"])
                    sid = self.token_factory["code"].key(user=_info["uid"], areq=areq)
                    dic = _info
                    dic["response_type"] = areq["response_type"].split(" ")
                else:
                    try:
                        dic = self._db[sid]
                    except KeyError:
                        dic = _info

                access_token = self.access_token(sid=sid, sinfo=dic)
                try:
                    at = dic["access_token"]
                except KeyError:
                    pass
                else:
                    if at:
                        self.access_token.invalidate(at)
            else:
                raise ExpiredToken()
        elif self.token_factory["refresh_token"] is None:
            raise WrongTokenType()
        elif self.token_factory["refresh_token"].valid(rtoken):
            if self.token_factory["refresh_token"].is_expired(rtoken):
                raise ExpiredToken()
            sid = self.token_factory["refresh_token"].get_key(rtoken)
            try:
                dic = self._db[sid]
            except KeyError:
                # Session is cleared, use the storage in token factory
                dic = self.token_factory["refresh_token"].token_storage[sid]
            access_token = self.access_token(sid=sid, sinfo=dic)

            try:
                at = dic["access_token"]
            except KeyError:
                pass
            else:
                if at:
                    self.access_token.invalidate(at)

            dic["access_token"] = access_token
        else:
            raise ExpiredToken()

        dic["access_token"] = access_token
        dic["token_type"] = "Bearer"  # nosec
        dic["refresh_token"] = rtoken
        dic["revoked"] = False
        self._db[sid] = dic
        return dic

    def is_valid(self, token, client_id=None):
        """
        Check validity of the given token.

        :param token: Access or refresh token
        :param client_id: Client ID, needed only for Refresh token
        """
        if token.startswith("Refresh_"):
            if self._refresh_db is None:
                return False
            return self._refresh_db.verify_token(client_id, token)

        try:
            typ, sid = self._get_token_type_and_key(token)
        except KeyError:
            return False

        if self.is_revoked(sid):
            return False

        _dic = self._db[sid]

        if typ == "A":
            if _dic["code"] != token:
                return False
            elif _dic["oauth_state"] != "authz":
                return False

        elif typ == "T":
            if _dic["access_token"] != token:
                return False

            if not self.access_token.valid(token):
                return False

        elif typ == "R":
            if self.token_factory["refresh_token"] is None:
                return False
            if not self.token_factory["refresh_token"].valid(token):
                return False
        return True

    def is_revoked(self, sid):
        try:
            return self[sid]["revoked"]
        except KeyError:
            return False

    def revoke_token(self, token):
        """
        Revoke access token.

        :param token: access token
        """
        _, sid = self._get_token_type_and_key(token)

        self.update(sid, "revoked", True)
        return True

    def revoke_refresh_token(self, rtoken):
        """
        Revoke refresh token.

        :param rtoken: Refresh token
        """
        if self._refresh_db:
            self._refresh_db.revoke_token(rtoken)
        elif self.token_factory["refresh_token"] is not None:
            self.token_factory["refresh_token"].invalidate(rtoken)

        return True

    def revoke_all_tokens(self, token):
        """
        Mark session as revoked but also explicitly revoke refresh token.

        :param token: access token
        """
        _, sid = self._get_token_type_and_key(token)

        try:
            rtoken = self._db[sid]["refresh_token"]
        except KeyError:
            pass
        else:
            self.revoke_refresh_token(rtoken)

        self.revoke_token(token)
        return True

    def get_client_id_for_session(self, sid):
        _dict = self._db[sid]
        return _dict["client_id"]

    def get_client_ids_for_uid(self, uid: str) -> List[str]:
        """Return client_ids for a given uid."""
        return self._db.get_client_ids_for_uid(uid)

    def get_verify_logout(self, uid: str) -> Optional[str]:
        """Return logout verification key for given uid."""
        return self._db.get_verified_logout(uid)

    def get_token_ids(self, uid: str) -> List[str]:
        """Return id_tokens for given uid."""
        return self._db.get_token_ids(uid)

    def set_verify_logout(self, uid: str) -> None:
        """Save the key that is used for logout verification."""
        for sid in self._db.get_by_uid(uid):
            _dict = self._db[sid]
            _dict["verified_logout"] = uuid.uuid4().urn

    def is_revoke_uid(self, uid: str) -> bool:
        """Return if the uid session has been revoked."""
        return self._db.is_revoke_uid(uid)

    def revoke_uid(self, uid: str) -> None:
        """Mark all sessions for the given uid as revoked."""
        for sid in self._db.get_by_uid(uid):
            self.update(sid, "revoked", True)

    def duplicate(self, sinfo):
        _dic = copy.copy(sinfo)
        areq = AuthorizationRequest().from_json(_dic["authzreq"])
        sid = self.token_factory["code"].key(user=_dic["sub"], areq=areq)

        _dic["code"] = self.token_factory["code"](sid=sid, sinfo=sinfo)
        _dic["code_used"] = False

        for key in [
            "access_token",
            "access_token_scope",
            "oauth_state",
            "token_type",
            "token_expires_at",
            "expires_in",
            "client_id_issued_at",
            "id_token",
            "oidreq",
            "refresh_token",
        ]:
            try:
                del _dic[key]
            except KeyError:
                pass

        self._db[sid] = _dic
        self._db.update(sid, "sub", _dic["sub"])

        return sid

    def read(self, token):
        (typ, key) = self._get_token_type_and_key(token)

        if typ != "T":  # not a access grant
            raise WrongTokenType("Not a grant token")

        return self._db[key]

    def get_by_sub(self, sub: str) -> List[str]:
        """Return session ids based on `sub` (external user identifier)."""
        return self._db.get_by_sub(sub)

    def make_smid(self, sid: str) -> str:
        """
        Create a session management ID.

        :param sid:
        :return: A session management ID
        """
        return hashlib.sha256(
            "{}{}".format(sid, self.sm_salt).encode("utf-8")
        ).hexdigest()

    def get(self, attr: str, val: Any) -> List[str]:
        """Return session ids based on attribute name and value."""
        return self._db.get(attr, val)

    def get_by_uid(self, uid: str) -> List[str]:
        """Return session ids (keys) based on `uid` (internal user identifier)."""
        return self._db.get_by_uid(uid)

    def get_uid_by_sub(self, sub: str) -> str:
        """Return session ids based on `sub` (external user identifier)."""
        return self._db.get_uid_by_sub(sub)

    def get_uid_by_sid(self, sub: str) -> str:
        """Return User id based on sub."""
        return self._db.get_uid_by_sid(sub)


def session_get(db, attr, val):
    """Return session ID based on attribute having value val."""
    if isinstance(db, (SessionBackend, SessionDB)):
        if attr == "uid":
            return db.get_by_uid(val)
        else:
            return db.get(attr, val)
    elif isinstance(db, dict):
        warnings.warn(
            "Using a regular dictionary is deprecated, please use `SessionDB` or `SessionBackend` instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        if attr == "uid":
            for _key, _val in db.items():
                if _val["authn_event"]["uid"] == val:
                    return _key
        else:
            for _key, _val in db.items():
                if _val.get(attr) == val:
                    return _key
        return None
    else:
        raise ValueError("Unknown session database type")


def session_extended_get(db, sub, attr, val):
    """Return session ID based on subject_id and attribute attr having value val."""
    if isinstance(db, (SessionBackend, SessionDB)):
        for sid in db.get_by_sub(sub):
            try:
                if db[sid][attr] == val:
                    return sid
            except KeyError:
                continue
        return None
    elif isinstance(db, dict):
        warnings.warn(
            "Using a regular dictionary is deprecated, please use `SessionDB` or `SessionBackend` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        for _key, _val in db.items():
            try:
                if _val["sub"] == sub and _val[attr] == val:
                    return _key
            except KeyError:
                continue
        return None
    else:
        raise ValueError("Unknown session database type")


def session_set(db, attr, val):
    if isinstance(db, SessionBackend):
        db[attr] = val
    elif isinstance(db, SessionDB):
        db._db.set(attr, val)
    elif isinstance(db, dict):
        warnings.warn(
            "Using a regular dictionary is deprecated, please use `SessionDB` or `SessionBackend` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        db[attr] = val
    else:
        raise ValueError("Unknown session database type")


def session_update(db, key, attr, val):
    if isinstance(db, SessionBackend):
        db.update(key, attr, val)
    elif isinstance(db, SessionDB):
        db._db.update(key, attr, val)
    elif isinstance(db, dict):
        warnings.warn(
            "Using a regular dictionary is deprecated, please use `SessionDB` or `SessionBackend` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        item = db[key]
        item[attr] = val
        db[key] = item
    else:
        raise ValueError("Unknown session database type")



import hashlib
import logging
import warnings

from jwkest import b64e

from oic import oauth2
from oic import unreserved
from oic.exception import AuthzError
from oic.exception import PyoidcError
from oic.extension.message import ClientRegistrationError
from oic.extension.message import ExtensionMessageFactory
from oic.oauth2.exception import Unsupported
from oic.oauth2.message import ErrorResponse
from oic.utils.http_util import SUCCESSFUL
from oic.utils.sanitize import sanitize
from oic.utils.settings import OauthClientSettings

logger = logging.getLogger(__name__)

__author__ = "roland"

# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------

RESPONSE2ERROR = {
    "ClientInfoResponse": [ClientRegistrationError],
    "ClientUpdateRequest": [ClientRegistrationError],
}

CC_METHOD = {"S256": hashlib.sha256, "S384": hashlib.sha384, "S512": hashlib.sha512}


class Client(oauth2.Client):
    def __init__(
        self,
        client_id=None,
        client_authn_method=None,
        keyjar=None,
        verify_ssl=None,
        config=None,
        message_factory=ExtensionMessageFactory,
        settings=None,
    ):
        self.settings = settings or OauthClientSettings()
        if verify_ssl is not None:
            warnings.warn(
                "`verify_ssl` is deprecated, please use `settings` instead if you need to set a non-default value.",
                DeprecationWarning,
                stacklevel=2,
            )
            self.settings.verify_ssl = verify_ssl
        super().__init__(
            client_id=client_id,
            client_authn_method=client_authn_method,
            keyjar=keyjar,
            config=config,
            message_factory=message_factory,
            settings=self.settings,
        )
        self.allow = {}
        self.request2endpoint.update(
            {
                "RegistrationRequest": "registration_endpoint",
                "ClientUpdateRequest": "clientinfo_endpoint",
                "TokenIntrospectionRequest": "introspection_endpoint",
                "TokenRevocationRequest": "revocation_endpoint",
            }
        )
        self.registration_response = None

    def construct_RegistrationRequest(
        self, request=None, request_args=None, extra_args=None, **kwargs
    ):
        if request is None:
            request = self.message_factory.get_request_type("registration_endpoint")
        if request_args is None:
            request_args = {}

        return self.construct_request(request, request_args, extra_args)

    def construct_ClientUpdateRequest(
        self, request=None, request_args=None, extra_args=None, **kwargs
    ):
        if request is None:
            request = self.message_factory.get_request_type("update_endpoint")
        if request_args is None:
            request_args = {}

        return self.construct_request(request, request_args, extra_args)

    def _token_interaction_setup(self, request_args=None, **kwargs):
        if request_args is None or "token" not in request_args:
            token = self.get_token(**kwargs)
            try:
                _token_type_hint = kwargs["token_type_hint"]
            except KeyError:
                _token_type_hint = "access_token"  # nosec

            request_args = {
                "token_type_hint": _token_type_hint,
                "token": getattr(token, _token_type_hint),
            }

        if "client_id" not in request_args:
            request_args["client_id"] = self.client_id
        elif not request_args["client_id"]:
            request_args["client_id"] = self.client_id

        return request_args

    def construct_TokenIntrospectionRequest(
        self, request=None, request_args=None, extra_args=None, **kwargs
    ):
        if request is None:
            request = self.message_factory.get_request_type("introspection_endpoint")
        request_args = self._token_interaction_setup(request_args, **kwargs)
        return self.construct_request(request, request_args, extra_args)

    def construct_TokenRevocationRequest(
        self, request=None, request_args=None, extra_args=None, **kwargs
    ):
        if request is None:
            request = self.message_factory.get_request_type("revocation_endpoint")
        request_args = self._token_interaction_setup(request_args, **kwargs)

        return self.construct_request(request, request_args, extra_args)

    def do_op(
        self,
        request,
        body_type="",
        method="GET",
        request_args=None,
        extra_args=None,
        http_args=None,
        response_cls=None,
        **kwargs,
    ):

        url, body, ht_args, _ = self.request_info(
            request, method, request_args, extra_args, **kwargs
        )

        if http_args is None:
            http_args = ht_args
        else:
            http_args.update(ht_args)

        resp = self.request_and_return(
            url, response_cls, method, body, body_type, http_args=http_args
        )

        return resp

    def do_client_registration(
        self,
        body_type="",
        method="GET",
        request_args=None,
        extra_args=None,
        http_args=None,
        **kwargs,
    ):
        request = self.message_factory.get_request_type("registration_endpoint")
        response_cls = self.message_factory.get_response_type("registration_endpoint")
        return self.do_op(
            request=request,
            body_type=body_type,
            method=method,
            request_args=request_args,
            extra_args=extra_args,
            http_args=http_args,
            response_cls=response_cls,
            **kwargs,
        )

    def do_client_read_request(
        self,
        body_type="",
        method="GET",
        request_args=None,
        extra_args=None,
        http_args=None,
        **kwargs,
    ):
        request = self.message_factory.get_request_type("update_endpoint")
        response_cls = self.message_factory.get_response_type("update_endpoint")
        return self.do_op(
            request=request,
            body_type=body_type,
            method=method,
            request_args=request_args,
            extra_args=extra_args,
            http_args=http_args,
            response_cls=response_cls,
            **kwargs,
        )

    def do_client_update_request(
        self,
        body_type="",
        method="PUT",
        request_args=None,
        extra_args=None,
        http_args=None,
        **kwargs,
    ):
        request = self.message_factory.get_request_type("update_endpoint")
        response_cls = self.message_factory.get_response_type("update_endpoint")
        return self.do_op(
            request=request,
            body_type=body_type,
            method=method,
            request_args=request_args,
            extra_args=extra_args,
            http_args=http_args,
            response_cls=response_cls,
            **kwargs,
        )

    def do_client_delete_request(
        self,
        body_type="",
        method="DELETE",
        request_args=None,
        extra_args=None,
        http_args=None,
        **kwargs,
    ):
        request = self.message_factory.get_request_type("delete_endpoint")
        response_cls = self.message_factory.get_response_type("delete_endpoint")
        return self.do_op(
            request=request,
            body_type=body_type,
            method=method,
            request_args=request_args,
            extra_args=extra_args,
            http_args=http_args,
            response_cls=response_cls,
            **kwargs,
        )

    def do_token_introspection(
        self,
        body_type="json",
        method="POST",
        request_args=None,
        extra_args=None,
        http_args=None,
        **kwargs,
    ):
        request = self.message_factory.get_request_type("introspection_endpoint")
        response_cls = self.message_factory.get_response_type("introspection_endpoint")
        return self.do_op(
            request=request,
            body_type=body_type,
            method=method,
            request_args=request_args,
            extra_args=extra_args,
            http_args=http_args,
            response_cls=response_cls,
            **kwargs,
        )

    def do_token_revocation(
        self,
        body_type=None,
        method="POST",
        request_args=None,
        extra_args=None,
        http_args=None,
        **kwargs,
    ):
        request = self.message_factory.get_request_type("revocation_endpoint")
        # There is no expected response, only the status code is important,
        # so do not use do_op().

        url, body, ht_args, _ = self.request_info(
            request, method, request_args, extra_args, **kwargs
        )

        if http_args is None:
            http_args = ht_args
        else:
            http_args.update(ht_args)

        resp = self.http_request(url, method, data=body, **http_args)

        if resp.status_code == 200:
            return 200
        if resp.status_code == 503:
            # Revoke failed, should retry later
            raise PyoidcError("Retry revocation later.")

        if 400 <= resp.status_code < 500:
            # check for error response
            try:
                err = ErrorResponse().deserialize(resp.text)
                try:
                    err.verify()
                except PyoidcError:
                    pass
                else:
                    return err
            except Exception:
                logger.exception(
                    "Failed to decode error response (%d) %s",
                    resp.status_code,
                    sanitize(resp.text),
                )

        return resp

    def add_code_challenge(self):
        try:
            cv_len = self.config["code_challenge"]["length"]
        except KeyError:
            cv_len = 64  # Use default

        code_verifier = unreserved(cv_len)
        _cv = code_verifier.encode("ascii")

        try:
            _method = self.config["code_challenge"]["method"]
        except KeyError:
            _method = "S256"

        try:
            _h = CC_METHOD[_method](_cv).digest()
            code_challenge = b64e(_h).decode("ascii")
        except KeyError:
            raise Unsupported("PKCE Transformation method:{}".format(_method))

        # TODO store code_verifier

        return (
            {"code_challenge": code_challenge, "code_challenge_method": _method},
            code_verifier,
        )

    def do_authorization_request(
        self,
        state="",
        body_type="",
        method="GET",
        request_args=None,
        extra_args=None,
        http_args=None,
        **kwargs,
    ):
        if "code_challenge" in self.config and self.config["code_challenge"]:
            _args, code_verifier = self.add_code_challenge()
            request_args.update(_args)

        oauth2.Client.do_authorization_request(
            self,
            state=state,
            body_type=body_type,
            method=method,
            request_args=request_args,
            extra_args=extra_args,
            http_args=http_args,
            **kwargs,
        )

    def store_registration_info(self, reginfo):
        self.registration_response = reginfo
        self.client_secret = reginfo["client_secret"]
        self.client_id = reginfo["client_id"]
        self.redirect_uris = reginfo["redirect_uris"]

    def handle_registration_info(self, response):
        if response.status_code in SUCCESSFUL:
            resp = self.message_factory.get_response_type(
                "registration_endpoint"
            )().deserialize(response.text, "json")
            self.store_response(resp, response.text)
            self.store_registration_info(resp)
        else:
            resp = ErrorResponse().deserialize(response.text, "json")
            try:
                resp.verify()
                self.store_response(resp, response.text)
            except Exception:
                raise PyoidcError("Registration failed: {}".format(response.text))

        return resp

    def register(self, url, **kwargs):
        """
        Register the client at an OP.

        :param url: The OPs registration endpoint
        :param kwargs: parameters to the registration request
        :return:
        """
        req = self.construct_RegistrationRequest(request_args=kwargs)

        headers = {"content-type": "application/json"}

        rsp = self.http_request(url, "POST", data=req.to_json(), headers=headers)

        return self.handle_registration_info(rsp)

    def parse_authz_response(self, query):
        aresp = self.parse_response(
            self.message_factory.get_response_type("authorization_endpoint"),
            info=query,
            sformat="urlencoded",
            keyjar=self.keyjar,
        )
        if isinstance(aresp, ErrorResponse):
            logger.info("ErrorResponse: %s" % sanitize(aresp))
            raise AuthzError(
                aresp.error  # type: ignore # Messages have no classical attrs
            )

        logger.info("Aresp: %s" % sanitize(aresp))

        return aresp
